import re
from typing import Optional
from typing_extensions import Self
from pydantic import Field, ValidationInfo, computed_field, conlist, model_validator, field_validator

from abc import ABC, abstractproperty, abstractstaticmethod

from ..utilities.well_known_keys import WellKnownKeys
from labfreed.validation import BaseModelWithValidationMessages, ValidationMessage, ValidationMsgLevel, hsegment_pattern, domain_name_pattern


class IDSegment(BaseModelWithValidationMessages):
    key:str|None = None
    value:str  
    
    @model_validator(mode="after")
    def validate_segment(self):
        key = self.key or ""
        value = self.value 
        
        # MUST be a valid hsegment according to RFC 1738, but without * (see PAC-ID Extension)
        # This means it must be true for both, key and value
        if not_allowed_chars := set(re.sub(hsegment_pattern, '', key)):
            self.add_validation_message(
                    source=f"id segment key {key}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f"{' '.join(not_allowed_chars)} must not be used. The segment key must be a valid hsegment",
                    highlight_pattern = key,
                    highlight_sub = not_allowed_chars
            )
        
        if not_allowed_chars := set(re.sub(hsegment_pattern, '', value)):
            self.add_validation_message(
                    source=f"id segment key {value}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f"{' '.join(not_allowed_chars)} must not be used. The segment key must be a valid hsegment",
                    highlight_pattern = value,
                    highlight_sub = not_allowed_chars
            )

        # Segment key SHOULD be limited to A-Z, 0-9, and -+..
        if not_recommended_chars := set(re.sub(r'[A-Z0-9-:+]', '', key)):
            self.add_validation_message(
                    source=f"id segment key {key}",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    msg=f"{' '.join(not_recommended_chars)} should not be used. Characters SHOULD be limited to upper case letters (A-Z), numbers (0-9), '-' and '+' ",
                    highlight_pattern = key,
                    highlight_sub = not_recommended_chars
                )
            
        # Segment key should be in Well know keys
        if key and key not in [k.value for k in WellKnownKeys]:
            self.add_validation_message(
                    source=f"id segment key {key}",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    msg=f"{key} is not a well known segment key. It is RECOMMENDED to use well-known keys.",
                    highlight_pattern = key,
                    highlight_sub=[key]
                )
            
            
        # Segment value SHOULD be limited to A-Z, 0-9, and -+..
        if not_recommended_chars := set(re.sub(r'[A-Z0-9-:+]', '', value)):
            self.add_validation_message(
                    source=f"id segment value {value}",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    msg=f"Characters {' '.join(not_recommended_chars)} should not be used., Characters SHOULD be limited to upper case letters (A-Z), numbers (0-9), '-' and '+' ",
                    highlight_pattern = value,
                    highlight_sub = not_recommended_chars
                )
            
        # Segment value SHOULD be limited to A-Z, 0-9, and :-+ for new designs.
        # this means that ":" in key or value is problematic
        if ':' in key:
            self.add_validation_message(
                    source=f"id segment key {key}",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    msg=f"Character ':' should not be used in segment key, since this character is used to separate key and value this can lead to undefined behaviour.",
                    highlight_pattern = key
                )
        if ':' in value:
            self.add_validation_message(
                    source=f"id segment value {value}",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    msg=f"Character ':' should not be used in segment value, since this character is used to separate key and value this can lead to undefined behaviour.",
                    highlight_pattern = value
                )
                
        return self
     
  

class PACID(BaseModelWithValidationMessages):
    issuer:str
    identifier: conlist(IDSegment, min_length=1) = Field(..., default_factory=list) # type: ignore # exclude=True prevents this from being serialized by Pydantic
        
        
    @model_validator(mode='after')
    def check_length(self) -> Self:
        l = 0
        for s in self.identifier:
            s:IDSegment = s
            if s.key:
                l += len(s.key)
                l += 1 # for ":"
            l += len(s.value)
        l += len(self.identifier) - 1 # account for "/" separating the segments
        
        if l > 256:
            self.add_validation_message(
                        source=f"identifier",
                        level = ValidationMsgLevel.ERROR,
                        msg=f'Identifier is {l} characters long, Identifier must not exceed 256 characters.'
                    )
        return self
       
    
    @model_validator(mode="after")
    def validate_issuer(self):
        if not re.fullmatch(domain_name_pattern, self.issuer):
            self.add_validation_message(
                    source="PAC-ID",
                    level = ValidationMsgLevel.ERROR,
                    highlight_pattern=self.issuer,
                    msg=f"Issuer must be a valid domain name. "
                )
         
        # recommendation that A-Z, 0-9, -, and . should be used
        if not_recommended_chars := set(re.sub(r'[A-Z0-9\.-]', '', self.issuer)):
            self.add_validation_message(
                    source="PAC-ID",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    highlight_pattern=self.issuer,
                    highlight_sub=not_recommended_chars,
                    msg=f"Characters {' '.join(not_recommended_chars)} should not be used. Issuer SHOULD contain only the characters A-Z, 0-9, -, and . "
                )
        return self
    
    
    @model_validator(mode='after')
    def check_identifier_segment_keys_are_unique(self) -> Self:
        keys = [s.key for s in self.identifier if s.key]
        duplicate_keys = [k for k in set(keys) if keys.count(k) > 1]
        if duplicate_keys:
            for k in duplicate_keys:
                self.add_validation_message(
                    source=f"identifier {k}",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    msg=f"Duplicate segment key {k}. This will probably lead to undefined behaviour",
                    highlight_pattern = k
                )
        return self
    
    
    def __str__(self):
        id_segments = ''
        for s in self.identifier:
            s:IDSegment = s
            if s.key:
                id_segments += f'/{s.key}:{s.value}'
            else:
                id_segments += f'/{s.value}'
    
        out = f"HTTPS://PAC.{self.issuer}{id_segments}"
        return out
    
    
    def serialize(self):
        return str(self)
    
    




import logging
import re

from .data_model import *
from labfreed.validation import LabFREEDValidationError

class TREX_Parser():
    def __init__(self, suppress_errors=False):
        self._suppress_errors = suppress_errors
    
    def parse_trex_str(self, trex_str, name=None) -> TREX:
        trex = _from_trex_string(trex_str, name=name)
        
        trex.print_validation_messages(trex_str)
        if not trex.is_valid() and not self._suppress_errors:
            raise LabFREEDValidationError(validation_msgs = trex.get_nested_validation_messages())
        
        return trex


def _from_trex_string(trex_str, name=None, enforce_type=True) -> TREX:
    if not trex_str:
        raise ValueError(f'T-REX must be a string of non zero length')

    # remove extension indicator. Precaution in case it is not done yet
    if trex_str[0]=="*":
        trex_str=trex_str[1:-1]
    # remove line breaks. for editing T-REXes it's more convenient to have them in, so one never knows 
    trex_str = trex_str.replace('\n','')
    
    d = re.match('((?P<name>.+)\$(?P<type>.+)/)?(?P<data>.+)', trex_str).groupdict()     
    if not d:
        raise ValueError('TREX is invalid.') 
    
    type = d.get('type')
    if not type:
        logging.warning('No type given. Assume its trex')
    elif type != 'TREX' and enforce_type:
        logging.error(f'Extension type {type} is not TREX. Aborting')
        raise ValueError(f'Extension type {type} is not TREX.')
    else:
        logging.warning('Extension type {type} is not TREX. Try anyways')
        
    s_name = d.get('name')
    if name and s_name:
        logging.warning(f'conflicting names given. The string contained {s_name}, method parameter was {name}. Method parameter wins.')
    elif not name and not s_name:
        raise ValueError('No extension name was given')
    elif s_name:
        name = s_name

    data = d.get('data')
    
    trex = TREX.from_spec_fields(name=name, data=data)

    return trex








from datetime import datetime
import os

import yaml
from labfreed.IO.parse_pac import PAC_Parser
from labfreed.PAC_ID_Resolver.resolver import PAC_ID_Resolver, load_cit
from labfreed.TREX.data_model import TREX
from labfreed.TREX.parse import TREX_Parser
from labfreed.DisplayNameExtension.DisplayNameExtension import DisplayNames


from labfreed.PAC_CAT.data_model import *
from labfreed.utilities.base36 import base36
from labfreed.utilities.utility_types import DataTable, Quantity, Unit

from labfreed.IO.generate_qr import save_qr_with_markers

if __name__ == "__main__":
    
    pac_with_ext = PAC_Parser().parse('HTTPS://PAC.METTORIUS.COM/-DR/maDFGAI/-MD/ABaCmG/ACG/PC:1234*SUM$TREX/A$T.B:X+B$T.X:ABC', suppress_errors=True)
    pac_with_ext.print_validation_messages()
    
    
    pac = PAC_Parser().parse('HTTPS://PAC.METTORIUS.COM/-DR/MADFGAI/-MD/ABCG/ACG/PC:1234')
    
    dir = os.path.dirname(__file__)
    p = os.path.join(dir, 'cit_mine.yaml')       
    cit = load_cit(p)

    services = PAC_ID_Resolver(cits=[cit]).resolve(pac)
    

    cat = PAC_CAT.from_pac_id(pac.pac_id)
    cat.identifier
    jsn = cat.model_dump_json(indent=2)
    print(jsn)
    
    
    
    save_qr_with_markers('HTTPS://PAC.METTORIUS.COM/-MD/ABCG/ACG') 
    
    
    table = DataTable(['DURATION', 'DATE', 'OK', 'COMMENT'])
    table.append([Quantity(value=1, unit=Unit(symbol='h', name='hour')), datetime.now(), True, 'FOO'])
    table.append([1.1,  datetime.now(), True, 'BAR'])
    table.append([1.3,  datetime.now(), False, 'BLUBB'])
    table.append([1.3,  datetime.now(), 1, 'BLUBB'])

    
    trex = TREX(name_='TEST')
    trex.update({'TABLE': table})
    trex.update({'TEMP': Quantity(value=100, unit=Unit(name='kelvin', symbol='K'))})
    trex.print_validation_messages()
    print(trex)
    
    d = trex.dict()
    
    
    
    
    parser = TREX_Parser()
    
    trex_str = 'A$T.T:ABCDE'
    trex = parser.parse_trex_str(trex_str, name="A")
    
    valid_strs = [
        'TIME$HUR:-25E2',
        'TIME$HUR:-25E-2',
        'TIME$HUR:-25.12E2',
        'TIME$HUR:-25.12E-2'
    ]
    for trex_str in valid_strs:
        trex = parser.parse_trex_str(trex_str, name="A")
        assert trex.data == trex_str
        v = trex.get_segment('TIME').to_python_type()
    
    # extension_interpreters = {
    #         'TREX': TREX,
    #         'N': DisplayNames
    # }

    # pacid_str = 'HTTPS://PAC.METTORIUS.COM/-DR/AB378/-MD/B-500/1235/-MS/AB/X:88/WWW/-MS/240:11/BB*E4BLEW6R5EVD7XMGHG11*A$HUR:25+B$CEL:99*BLUBB$TREX/A$HUR:25+B$CEL:99'
    # pac = PAC_Parser(extension_interpreters).parse_pac(pacid_str)
    # a=1
    



from datetime import date, datetime, time
import logging
import re
from collections import Counter
from typing import Annotated, Literal


from pydantic import PrivateAttr, RootModel, ValidationError, field_validator, model_validator, Field
from labfreed.TREX.unece_units import unece_unit, unece_unit_codes, unece_units, unit_name, unit_symbol
from labfreed.utilities.utility_types import DataTable, Quantity, Unit, unece_unit_code_from_quantity
from labfreed.validation import BaseModelWithValidationMessages, ValidationMsgLevel
from abc import ABC,  abstractmethod

from labfreed.PAC_ID.extensions import Extension
from labfreed.utilities.base36 import base36, to_base36, from_base36

        

class TREX_Segment(BaseModelWithValidationMessages, ABC):
    key: str
    
    @model_validator(mode='after')
    def validate_key(self):
        if not_allowed_chars := set(re.sub(r'[A-Z0-9\.-]', '', self.key)):
            self.add_validation_message(
                source=f"TREX segment key {self.key}",
                type="Error",
                msg=f"Segment key contains invalid characters: {','.join(not_allowed_chars)}",
                highlight_pattern = f'{self.key}$',
                highlight_sub=not_allowed_chars
            )
        return self
    
    @abstractmethod
    def serialize_for_trex(self):
        raise NotImplementedError("Subclasses must implement 'serialize_as_trex()' method")
    
    # @abstractmethod
    # def to_python_type(self):
    #      raise NotImplementedError("Subclasses must implement 'to_python_type()' method")
     
    # @abstractmethod
    # def from_python_type(self):
    #      raise NotImplementedError("Subclasses must implement 'from_python_type()' method")
    
    


class ValueMixin(BaseModelWithValidationMessages, ABC): 
    value:str 
    
    def serialize_for_trex(self):
        return self.value
        
    # @abstractclassmethod
    # def from_python_type(cls, v):
    #     ...
        
    @abstractmethod
    def value_to_python_type(self):
        ...
        
        
class NumericValue(ValueMixin):
    @field_validator('value', mode='before')
    @classmethod
    def from_python_type(cls, v:str| int|float):
        if isinstance(v, str):
            return v
        return str(v)
        
    @model_validator(mode='after')
    def validate(self):
        value = self.value
        if not_allowed_chars := set(re.sub(r'[0-9\.\-E]', '', value)):
            self.add_validation_message(
                source=f"TREX numeric value {value}",
                type="Error",
                msg=f"Characters {','.join(not_allowed_chars)} are not allowed in quantity segment. Base36 encoding only allows A-Z0-9",
                highlight_pattern = f'{value}',
                highlight_sub=not_allowed_chars
            )
        if not re.fullmatch(r'-?\d+(\.\d+)?(E-?\d+)?', value):
            self.add_validation_message(
                source=f"TREX numeric value {value}",
                type="Error",
                msg=f"{value} cannot be converted to number",
                highlight_pattern = f'{value}'               
            )
        return self
    
    def value_to_python_type(self) -> str:
        v = float(self.value)  
        if not '.' in self.value and not 'E' in self.value: 
            return int(v)
        else:
            return v


class DateValue(ValueMixin):
    _date_time_dict:dict|None = PrivateAttr(default=None)
    @field_validator('value', mode='before')
    @classmethod
    def from_python_type(cls, v:str| date|time|datetime):
        if isinstance(v, str):
            return v
        
        sd = ""
        st = ""
        if isinstance(v, date) or isinstance(v, datetime):
            sd = v.strftime('%Y%m%d')
        if isinstance(v, time) or isinstance(v, datetime):
            if v.microsecond:
                st = v.strftime("T%H%M%S.") + f"{v.microsecond // 1000:03d}"
            elif v.second:
                st = v.strftime("T%H%M%S")
            else:
                st = v.strftime("T%H%M")
                         
        return sd + st
    
    @model_validator(mode='after')
    def validate(self):
        pattern:str = r'((?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2}))?(T(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})?(\.(?P<millisecond>\d{3}))?)?'
        value=self.value
        if not re.fullmatch(pattern, value):
            self.add_validation_message(
                source=f"TREX date value {value}",
                type="Error",
                msg=f'{value} is not in a valid format. Valid format for date: YYYYMMDD; Valid for time: THHMM, THHMMSS, THHMMSS.SSS; Datetime any combination of valid date and time',
                highlight_pattern = f'{value}'
            )
            return self
            
        matches = re.match(pattern, value)       
        d = matches.groupdict()  
        d = {k: int(v) for k,v in d.items() if v }
        if 'millisecond' in d.keys():
            ms = d.pop('millisecond')
            d.update({'microsecond': ms * 1000})
        try:
            if d.get('year'): # input is only a time
                datetime(**d)
            else:
                time(**d)
        except ValueError as e:
            self.add_validation_message(
                source=f"TREX date value {value}",
                type="Error",
                msg=f'{value} is no valid date or time.',
                highlight_pattern = f'{value}'
            )
            
        self._date_time_dict = d
        return self
    
    def value_to_python_type(self) -> str:
        d = self._date_time_dict
        if d.get('year') and d.get('hour'): # input is only a time
            return datetime(**d)
        elif d.get('year'):
            return date(**d)
        else:
            return time(**d)
    
   
    

class BoolValue(ValueMixin):
    @field_validator('value', mode='before')
    @classmethod
    def from_python_type(cls, v:str| bool):
        if isinstance(v, str):
            return v
        
        return 'T' if v else 'F'

    @model_validator(mode='after')
    def validate(self):
        if not self.value in ['T', 'F']:
            self.add_validation_message(
                source=f"TREX boolean value {self.value}",
                level= ValidationMsgLevel.ERROR,
                msg=f'{self.value} is no valid boolean. Must be T or F',
                highlight_pattern = f'{self.value}',
                highlight_sub=[c for c in self.value]
            )
        return self
    
    def value_to_python_type(self) -> str:
        if self.value == 'T':
            return True
        elif self.value == 'F':
            return False
        else:
            Exception(f'{self} is not valid boolean. That really should not have been possible -- Contact the maintainers of the library')
            
          
class AlphanumericValue(ValueMixin):
    @field_validator('value', mode='before')
    @classmethod
    def from_python_type(cls, v:str):
        return v
        
    @model_validator(mode='after')
    def validate(self):
        if re.match(r'[a-z]', self.value):
            self.add_validation_message(
                    source=f"TREX value {self.value}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Lower case characters are not allowed.",
                    highlight_pattern = self.value
            )
            
        if not_allowed_chars := set(re.sub(r'[A-Z0-9\.-]', '', self.value)):
            self.add_validation_message(
                    source=f"TREX value {self.value}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Characters {','.join(not_allowed_chars)} are not allowed in alphanumeric segment",
                    highlight_pattern = self.value,
                    highlight_sub=not_allowed_chars
            )
        return self
           
    def value_to_python_type(self) -> str:
        return self.value
    
    

class TextValue(ValueMixin):
    @field_validator('value', mode='before')
    @classmethod
    def from_python_type(cls, v:base36|str):
        if isinstance(v, str):
            logging.info('Got str for text value > converting to base36')
            return to_base36(v).root
        else:
            return v.root
        
    @model_validator(mode='after')
    def validate(self):
        if not_allowed_chars := set(re.sub(r'[A-Z0-9]', '', self.value)):
            self.add_validation_message(
                    source=f"TREX value {self.value}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Characters {','.join(not_allowed_chars)} are not allowed in text segment. Base36 encoding only allows A-Z0-9",
                    highlight_pattern = self.value,
                    highlight_sub=not_allowed_chars
            )
        return self
              
    def value_to_python_type(self) -> str:
        decoded = from_base36(self.value)
        return decoded
    
    
class BinaryValue(ValueMixin):
    @field_validator('value', mode='before')
    @classmethod
    def from_python_type(cls, v:base36|str):
        if isinstance(v, str):
            return v
        else:
            return v.root
        
    @model_validator(mode='after')
    def validate(self):
        if not_allowed_chars := set(re.sub(r'[A-Z0-9]', '', self.value)):
           self.add_validation_message(
                    source=f"TREX value {self.value}",
                    tlevel= ValidationMsgLevel.ERROR,
                    msg=f"Characters {','.join(not_allowed_chars)} are not allowed in text segment. Base36 encoding only allows A-Z0-9",
                    highlight_pattern = self.value,
                    highlight_sub=not_allowed_chars
            )
        return self
              
    def value_to_python_type(self) -> bytes:
        decoded = bytes(from_base36(self))
        return decoded
    
    
class ErrorValue(ValueMixin):
    @model_validator(mode='after')
    def validate(self):
        if not_allowed_chars := set(re.sub(r'[A-Z0-9\.-]', '', self.value)):
            self.add_validation_message(
                    source=f"TREX value {self.value}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Characters {','.join(not_allowed_chars)} are not allowed in error segment",
                    highlight_pattern = self.value,
                    highlight_sub=not_allowed_chars
            )
        return self
       
    
    def value_to_python_type(self) -> str:
        return self.value



        
class ValueSegment(TREX_Segment, ValueMixin, ABC):
    type:str
    
    @model_validator(mode='after')
    def validate_type(self):
        valid_types = valid_types = unece_unit_codes() + ['T.D', 'T.B', 'T.A', 'T.T', 'T.X', 'E']
        if not self.type in valid_types:
            self.add_validation_message(
                    source=f"TREX value segment {self.key}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Type {self.type} is invalid. Must be 'T.D', 'T.B', 'T.A', 'T.T', 'T.X', 'E' or a UNECE unit",
                    highlight_pattern = self.type
            )
        return self
    
    # @classmethod
    # def get_subclass(cls, type:str, key:str, value:str):
    #     match type:
    #         case 'T.D':
    #             model = DateSegment(key=key, value=value, type=type)
    #         case 'T.B':
    #             model = BoolSegment(key=key, value=value, type=type)
    #         case 'T.A':
    #             model = AlphanumericSegment(key=key, value=value, type=type)
    #         case 'T.T':
    #             model = TextSegment(key=key, value=value, type=type)
    #         case 'T.X':
    #             model = BinarySegment(key=key, value=value, type=type)
    #         case 'E':
    #             model = ErrorSegment(key=key, value=value, type=type)
    #         case _:
    #             model = NumericSegment(value=value, key=key, type=type)
                    
    #     return model    
    
    
    def serialize_for_trex(self) -> str:
        return f'{self.key}${self.type}:{self.value}'
    
    def to_python_type(self):
        return self.value_to_python_type()
    
    

    
    

    
class NumericSegment(ValueSegment, NumericValue):
    type: str    
    
    def to_python_type(self):
        unit = unece_unit(self.type)
        out = Quantity(value=self.value_to_python_type(), unit=Unit(name=unit_name(unit), symbol=unit_symbol(unit)))
        return out
    
class DateSegment(ValueSegment, DateValue):
    type: Literal['T.D'] = Field('T.D', frozen=True)
          
class BoolSegment(ValueSegment, BoolValue):
    type: Literal['T.B']  = Field('T.B', frozen=True) 
        
class AlphanumericSegment(ValueSegment, AlphanumericValue):
    type: Literal['T.A'] = Field('T.A', frozen=True)
    
class TextSegment(ValueSegment, TextValue):
    type: Literal['T.T'] = Field('T.T', frozen=True)
    
class BinarySegment(ValueSegment, BinaryValue):
    type: Literal['T.X'] = Field('T.X', frozen=True)
    
class ErrorSegment(ValueSegment, ErrorValue):
    type: Literal['E'] = Field('E', frozen=True)
                        
        
    
class ColumnHeader(BaseModelWithValidationMessages):
    key:str
    type:str
               
    @model_validator(mode='after')     
    def validate_key(self):
        if not_allowed_chars := set(re.sub(r'[A-Z0-9\.-]', '', self.key)):
            self.add_validation_message(
                source=f"TREX table column {self.key}",
                level= ValidationMsgLevel.ERROR,
                msg=f"Column header key contains invalid characters: {','.join(not_allowed_chars)}",
                highlight_pattern = f'{self.key}$',
                highlight_sub=not_allowed_chars
            )
        return self
    
    @model_validator(mode='after')
    def validate_type(self):
        valid_types = unece_unit_codes() + ['T.D', 'T.B', 'T.A', 'T.T', 'T.X', 'E']
        if not self.type in valid_types:
            self.add_validation_message(
                    source=f"TREX table column {self.key}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Type '{self.type}' is invalid. Must be 'T.D', 'T.B', 'T.A', 'T.T', 'T.X', 'E' or a UNECE unit",
                    highlight_pattern = self.type
            )
        return self

class TableRow(RootModel[list[ValueMixin]]):    
    def serialize_for_trex(self):
        return ':'.join([e.serialize_for_trex() for e in self.root])
    
    def __len__(self):
        return len(self.root)
    
    def __iter__(self):
        return iter(self.root)
  
class TREX_Table(TREX_Segment):
    column_headers: list[ColumnHeader]
    data: list[TableRow]
    
    @property
    def column_names(self):
        return [h.key for h in self.column_headers]
    
    @property
    def column_types(self):
        return [h.type for h in self.column_headers]
    
    @model_validator(mode='after')
    def validate_sizes(self):
        sizes = [len(self.column_headers)]
        sizes.extend( [ len(row) for row in self.data ] ) 
        most_common_len, count = Counter(sizes).most_common(1)[0]        
        
        if len(self.column_headers) != most_common_len:
            self.add_validation_message(
                    source=f"Table {self.key}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Size mismatch: Table header contains {self.column_names} keys, while most rows have {most_common_len}",
                    highlight_pattern = self.key
            )
            expected_row_len = most_common_len
        else:
            expected_row_len = len(self.column_headers)
            

        for i, row in enumerate(self.data):
            if len(row) != expected_row_len:
                self.add_validation_message(
                    source=f"Table {self.key}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Size mismatch: Table row {i} contains {len(row)} elements. Expected size is {expected_row_len}",
                    highlight_pattern = row.serialize_for_trex()
                )
        return self
    
    @model_validator(mode='after')
    def validate_data_types(self):
        expected_types = self.column_types
        i = 0
        for row in self.data:
            for e, t_expected, nm in zip(row, expected_types, self.column_names):
                try:
                    match t_expected:
                        case 'T.D':
                            assert isinstance(e, DateValue)
                        case 'T.B':
                            assert isinstance(e, BoolValue)
                        case 'T.A':
                            assert isinstance(e, AlphanumericValue)
                            
                        case 'T.T':
                            assert isinstance(e, TextValue)
                        case 'T.X':
                            assert isinstance(e, BinaryValue)
                        case 'E':
                            assert isinstance(e, ErrorValue)
                        case _:
                            assert isinstance(e, NumericValue)
                except AssertionError:
                    self.add_validation_message(
                        source=f"Table {self.key}",
                        level= ValidationMsgLevel.ERROR,
                        msg=f"Type mismatch: Table row {i}, column {nm} is of wrong type. According to the header it should be {t_expected}",
                        highlight_pattern = row.serialize_for_trex(),
                        highlight_sub=[c for c in e.value]
                    )
                    
                if msg := e.get_errors():
                    for m in msg:
                        self.add_validation_message(
                            source=f"Table {self.key}",
                            level= ValidationMsgLevel.ERROR,
                            msg=m.problem_msg,
                            highlight_pattern = row.serialize_for_trex(),
                            highlight_sub=[c for c in e.value]
                        )
                i += 1
                
            
    def _get_col_index(self, col:str|int):
        if isinstance(col, str):
            col_index = self.column_names.index(col)
        elif isinstance(col, int):
            col_index = col
        else:
            raise TypeError(f"Column must be specified as string or int: {col.__name__}")
        return col_index
    
    
    
    def serialize_for_trex(self):
        header = ':'.join([f'{h.key}${h.type}' for h in self.column_headers])        
        data = '::'.join([r.serialize_for_trex() for r in self.data])
        s = f'{self.key}$${header}::{data}'
        return s
    
    
    def to_python_type(self):
        table = DataTable([ch.key for ch in self.column_headers])
        for row in self.data:
            r = []
            for e, h in zip(row, self.column_headers):
                if isinstance(e, NumericValue):
                    u = unece_unit(h.type)
                    unit = Unit(name=u.get('name'), symbol=u.get('symbol'))
                    r.append(Quantity(value=e.value, unit=unit))
                else:
                    r.append(e.value_to_python_type())
            table.append(r)
        return table
        
        
        
    def n_rows(self) -> int:
        return len(self.data)
    
    def n_cols(self) -> int:
        return len(self.column_headers)
      
    def row_data(self, row:int) -> list:
        out = self.data[row]
        return out
    
    
    def column_data(self, col:str|int) -> list:
        col_index = self._get_col_index(col)
        type = self.column_headers[col_index].type
        out =  [row[col_index] for row in self.data]
        return out
     
     
    def cell_data(self, row:int, col:str|int):
        try:
            col_index = self._get_col_index(col)
            value = self.data[row][col_index]
        except ValueError:
            logging.warning(f"row {row}, column {col} not found")
            return None 
        return value
   



class TREX(Extension, BaseModelWithValidationMessages):
    name_:str
    segments: list[TREX_Segment] = Field(default_factory=list)
       
    @property
    def name(self)->str:
        return self.name_
    
    @property
    def type(self)->str:
        return 'TREX'
    
    @property
    def data(self)->str:
        seg_strings = list()
        for s in self.segments:
            seg_strings.append(s.serialize_for_trex())
        s_out = '+'.join(seg_strings)
        return s_out
    
       
    def get_segment(self, segment_key:str) -> TREX_Segment:
        s = [s for s in self.segments if s.key == segment_key]
        if s:
            return s[0]
        else:
            return None
        
        
    def update(self, segments: dict[str, Quantity|datetime|time|date|bool|str|base36|DataTable] ):
        for k, v in segments.items():
            if isinstance(v, bool):
                self.segments.append(BoolSegment(key=k, value=v))
            elif isinstance(v, Quantity):
                unece_code = unece_unit_code_from_quantity(v)
                self.segments.append(NumericSegment(key=k, value=v.value, type=unece_code))
            elif isinstance(v, (int, float)):
                self.segments.append(NumericSegment(key=k, value=v, type='C63'))  # unitless
            elif isinstance(v, (datetime, time, date)):
                self.segments.append(DateSegment(key=k, value=v))
            elif isinstance(v, str):
                if re.fullmatch(r'[A-Z0-9\-\.]*', v):
                    self.segments.append(AlphanumericSegment(key=k, value=v))
                else:
                    v = to_base36(v)
                    self.segments.append(TextSegment(key=k, value=v))
            elif isinstance(v, base36):
                self.segments.append(TextSegment(key=k, value=v))
            elif isinstance(v, DataTable):
                v:DataTable = v
                headers = list()
                for nm, rt in zip(v.col_names, v.row_template):
                    if isinstance(rt, bool): # must come first otherwise int matches the bool
                        t = 'T.B'
                    elif isinstance(rt, Quantity):
                        unece_code = unece_unit_code_from_quantity(rt)
                        t = unece_code
                    elif isinstance(rt, (datetime, time, date)):
                        t = 'T.D'     
                    elif isinstance(rt, str):
                        if re.fullmatch(r'[A-Z0-9\-\.]*', rt):
                            t = 'T.A'
                        else:
                            v = to_base36(rt)
                            t = 'T.X'
                    elif isinstance(rt, base36):
                        t = 'T.X'
                    
                    headers.append(ColumnHeader(key=nm, type=t))
                data = []
                for row in v:
                    r = []
                    for e in row:
                        if isinstance(e, bool): # must come first otherwise int matches the bool
                            r.append(BoolValue(value=e))
                        elif isinstance(e, Quantity):
                            r.append(NumericValue(value=e.value))
                        elif isinstance(e, (int, float)):
                            r.append(NumericValue(value=e))
                        elif isinstance(e, (datetime, time, date)):
                            r.append(DateValue(value=e))
                        elif isinstance(e, str):
                            if re.fullmatch(r'[A-Z0-9\-\.]*', e):
                                r.append(AlphanumericValue(value=e))
                            else:
                                e = to_base36(e)
                                r.append(TextValue(value=e))
                        elif isinstance(e, base36):
                            r.append(TextValue(value=e))
                    data.append(r)
                    
                self.segments.append(TREX_Table(key=k, column_headers=headers, data=data))
        return self
                   
                
    def dict(self):
        return {s.key: s.to_python_type() for s in self.segments}
            
        
        
    @field_validator('segments')
    @classmethod
    def validate_segments(cls, segments):
        segment_keys = [s.key for s in segments]
        duplicates = [item for item, count in Counter(segment_keys).items() if count > 1]
        if duplicates:
            raise ValueError(f"Duplicate segment keys: {','.join(duplicates)}")
        return segments
      
    
        
    @staticmethod
    def from_spec_fields(*, name, data, type='TREX'):
        segment_strings = data.split('+')
        out_segments = list()
        for s in segment_strings:
            # there are only two valid options. The segment is a scalar or a table. 
            # Constructors do the parsing anyways and raise exceptions if invalid data
            # try both options and then let it fail
            segment = _deserialize_table_segment_from_trex_segment_str(s)
            if not segment:
                segment = _deserialize_value_segment_from_trex_segment_str(s)
            if not segment:
                raise ValueError('TREX contains neither valid value segment nor table')
                
            out_segments.append(segment)
        trex = TREX(name_= name, segments=out_segments)
    
        return trex
    
    
def _deserialize_value_segment_from_trex_segment_str(trex_segment_str) -> ValueSegment:
    #re_scalar_pattern = re.compile(f"(?P<name>[\w\.-]*?)\$(?P<unit>[\w\.]*?):(?P<value>.*)")
    re_scalar_pattern = re.compile(f"(?P<name>.+?)\$(?P<unit>.+?):(?P<value>.+)")
    matches = re_scalar_pattern.match(trex_segment_str)
    if not matches:
        return None
    
    key, type_, value = matches.groups()
    
    match type_:
        case 'T.D':
            out = DateSegment(key=key, value=value, type=type_)
        case 'T.B':
            out = BoolSegment(key=key, value=value, type=type_)
        case 'T.A':
            out = AlphanumericSegment(key=key, value=value, type=type_)
        case 'T.T':
            out = TextSegment(key=key, value=base36(value), type=type_)  # prevent repeated conversion from str to base36 and make explict that when parsing we assume the string tpo be base36 already
        case 'T.X':
            out = BinarySegment(key=key, value=base36(value), type=type_) # prevent repeated conversion from str to base36 and make explict that when parsing we assume the string tpo be base36 already
        case 'E':
            out = ErrorSegment(key=key, value=value, type=type_)
        case _:
            out = NumericSegment(value=value, key=key, type=type_)
                    
    return out    
    


def _deserialize_table_segment_from_trex_segment_str(trex_segment_str) -> TREX_Table:
    # re_table_pattern = re.compile(f"(?P<tablename>[\w\.-]*?)\$\$(?P<header>[\w\.,\$:]*?)::(?P<body>.*)")
    # re_col_head_pattern = re.compile(f"(?P<name>[\w\.-]*?)\$(?P<unit>[\w\.]*)")
    re_table_pattern = re.compile(r"(?P<tablename>.+?)\$\$(?P<header>.+?)::(?P<body>.+)")
    
    matches = re_table_pattern.match(trex_segment_str) 
    if not matches:
        return None
    name, header, body = matches.groups()
    
    column_headers_str = header.split(':')
    
    headers = []
    for colum_header in column_headers_str:
         ch = colum_header.split('$')
         col_key = ch[0]
         col_type = ch[1] if len(ch) > 1 else ''
         headers.append(ColumnHeader(key=col_key, type=col_type))
    
    data = [row.split(':') for row in body.split('::') ]
    col_types = [h.type for h in headers]
    # convert to correct value types
    data_with_types = [[str_to_value_type(c,t) for c, t in zip(r, col_types)] for r in data]
    data = [ TableRow(r) for r in data_with_types]
             
    out = TREX_Table(column_headers=headers, data=data_with_types, key=name)
    return out
        

def str_to_value_type(s:str, t:str):
    match t:
        case 'T.D': v = DateValue(value=s)
        case 'T.B': v = BoolValue(value=s)
        case 'T.A': v = AlphanumericValue(value=s)
        case 'T.T': v = TextValue(value=base36(s))
        case 'T.X': v = BinaryValue(value=s)
        case 'E'  : v = ErrorValue(value=s)
        case _    : v = NumericValue(value=s)         
    return v   
    

    
class TREX_Struct(TREX_Segment):
    """Struct is a special interpretation of a T-REX Table with one row"""
    wrapped_table:TREX_Table
    
    @property
    def segment_name_(self):
        return self.wrapped_table.key
    
    @field_validator('wrapped_table')
    def validate_table(table):
        if len(table.data) != 1:
            raise ValidationError("Too many input rows. Struct can only have one row")
        return table
                
    def get(self, key):
        return self.wrapped_table.cell_data(0, key)
    
    def keys(self):
        return self.wrapped_table.col_names
    
    
    





# import built ins
import os
''' 
### Parse a simple PAC-ID 
'''
from labfreed.IO.parse_pac import PAC_Parser


# Parse the PAC-ID
pac_str = 'HTTPS://PAC.METTORIUS.COM/-MD/bal500/@1234'
pac_id = PAC_Parser().parse(pac_str).pac_id

# Check validity of this PAC-ID
is_valid = pac_id.is_valid()
print(f'PAC-ID is valid: {is_valid}')

''' 
### Show recommendations:
Note that the PAC-ID -- while valid -- uses characters which are not recommended (results in larger QR code).
There is a nice function to highlight problems
'''
pac_id.print_validation_messages(target='markdown')

'''
### Save as QR Code
'''
from labfreed.IO.generate_qr import save_qr_with_markers

save_qr_with_markers(pac_str, fmt='png')

'''
### PAC-CAT
'''
from labfreed.PAC_CAT.data_model import PAC_CAT
pac_str = 'HTTPS://PAC.METTORIUS.COM/-DR/XQ908756/-MD/bal500/@1234'
pac_id = PAC_Parser().parse(pac_str).pac_id
if isinstance(pac_id, PAC_CAT):
    pac_id.print_categories()



''' 
### Parse a PAC-ID with extensions
PAC-ID can have extensions. Here we parse a PAC-ID with attached display names and summary.
'''
pac_str = 'HTTPS://PAC.METTORIUS.COM/-MD/BAL500/1234*N$N/WM633OV3E5DGJW2BEG0PDM1EA7*SUM$TREX/WEIGHT$GRM:67.89'
pac_id = PAC_Parser().parse(pac_str)

# Display Name
display_names = pac_id.get_extension('N') # display name has name 'N'
print(display_names)

''''''
# TREX
trexes = pac_id.get_extension_of_type('TREX')
trex = trexes[0] # there could be multiple trexes. In this example there is only one, though
v = trex.get_segment('WEIGHT').to_python_type() 
print(f'WEIGHT = {v}')



''' 
### Create a PAC-ID with Extensions

#### Create PAC-ID
'''
from labfreed.PAC_ID.data_model import PACID, IDSegment
from labfreed.utilities.well_known_keys import WellKnownKeys

pac_id = PACID(issuer='METTORIUS.COM', identifier=[IDSegment(key=WellKnownKeys.SERIAL, value='1234')])
pac_str = pac_id.serialize()
print(pac_str)


''' 
#### Create a TREX 
TREX can conveniently be created from a python dictionary.
Note that utility types for Quantity (number with unit) and table are needed
'''
from datetime import datetime
from labfreed.TREX.data_model import TREX
from labfreed.utilities.utility_types import Quantity, DataTable, Unit

# Create TREX
trex = TREX(name_='DEMO') 
# Add value segments of different type
trex.update(   
                    {
                        'STOP': datetime(year=2024,month=5,day=5,hour=13,minute=6),
                        'TEMP': Quantity(value=10.15, unit=Unit(name='kelvin', symbol='K')),
                        'OK':False,
                        'COMMENT': 'FOO',
                        'COMMENT2':'£'
                    }
)

# Create a table
table = DataTable(['DURATION', 'Date', 'OK', 'COMMENT'])
table.append([Quantity(value=1, unit=Unit(symbol='h', name='hour')), datetime.now(), True, 'FOO'])
table.append([                                                 1.1,  datetime.now(), True, 'BAR'])
table.append([                                                 1.3,  datetime.now(), False, 'BLUBB'])
#add the table to the trex
trex.update({'TABLE': table})

# Validation also works the same way for TREX
trex.print_validation_messages(target='markdown')
''''''
# there is an error. 'Date' uses lower case. Lets fix it
d = trex.dict()
d['TABLE'].col_names[1] = 'DATE'
trex = TREX(name_='DEMO') 
trex.update(d)

''' 
#### Combine PAC-ID and TREX and serialize
'''
from labfreed.IO.parse_pac import PACID_With_Extensions

pac_with_trex = PACID_With_Extensions(pac_id=pac_id, extensions=[trex])
pac_str = pac_with_trex.serialize()
print(pac_str)





'''
## PAC-ID Resolver
'''
from labfreed.PAC_ID_Resolver.resolver import PAC_ID_Resolver, load_cit
# Get a CIT
dir = os.path.dirname(__file__)
p = os.path.join(dir, 'cit_mine.yaml')       
cit = load_cit(p)

# validate the CIT
cit.is_valid()
cit.print_validation_messages()

''''''
# resolve a pac id
service_groups = PAC_ID_Resolver(cits=[cit]).resolve(pac_with_trex)
for sg in service_groups:
    sg.update_states()
    sg.print()
    
5



import pytest
from labfreed.PAC_ID.extensions import Extension
from labfreed.IO.parse_pac import PAC_Parser


valid_base = "HTTPS://PAC.METTORIUS.COM/"
valid_standard_segments = "-MD/240:B-800/21:12345"
valid_dummy_extension = "DUMMY$MYTYPE/DUMMYDATA"

parser = PAC_Parser()

# Extensions
def test_valid_extensions():
    pac = parser.parse(valid_base + valid_standard_segments + "*name1$t1/data1*name2$t2/data2")
    extensions = pac.extensions
    ext: Extension = extensions[0]
    assert ext.name == 'name1'
    assert ext.type == 't1'
    assert ext.data == 'data1'
    
    ext: Extension = extensions[1]
    assert ext.name == 'name2'
    assert ext.type == 't2'
    assert ext.data == 'data2'
    
def test_known_extension_types_are_parsed():
    class ExtensionMockType(Extension):
        @property
        def name(self)->str:
            return 'name__foo'
        
        @property
        def type(self)->str:
            return 'type__foo'
        
        @property
        def data(self)->str:
            return 'data__foo'
        
        @staticmethod
        def from_spec_fields(name, type, data):
            return ExtensionMockType()
        
    extension_interpreters = {
            'KNOWN_EXTENSION': ExtensionMockType,
    }  
    parser_with_known_extension = PAC_Parser(extension_interpreters)
    pac = parser_with_known_extension.parse(valid_base + valid_standard_segments + "*name1$KNOWN_EXTENSION/data1")
    extensions= pac.extensions
    ext: Extension = extensions[0]
    assert isinstance(ext, ExtensionMockType)
    assert ext.name == 'name__foo'
    assert ext.type == 'type__foo'
    assert ext.data == 'data__foo'
    
    
def test_imply_display_name_and_summary_extension():
    pac = parser.parse(valid_base + valid_standard_segments + "*data1*data2")
    extensions = pac.extensions
    ext: Extension = extensions[0]
    assert ext.name == 'N'
    assert ext.type == 'N'
    
    ext: Extension = extensions[1]
    assert ext.name == 'SUM'
    assert ext.type == 'TREX'
    
def test_stop_imply_extensions_after_explicit():
    with pytest.raises(Exception):
        pac = parser.parse(valid_base + valid_standard_segments + "*N$T/data1*data2")
        pac.extensions
        

def test_extension_parsing():
    s = '*NAME$MYFORMAT/AUGDSJGTZFRDGJHDSFRTZGHJAAUTZSGADT*NAME$ANOTHERFORMAT/BLUBBER'
    extensions = parser._parse_extensions(s)
    ext: Extension = extensions[0]
    assert ext.name == 'NAME'
    assert ext.type == 'MYFORMAT'
    assert ext.data == 'AUGDSJGTZFRDGJHDSFRTZGHJAAUTZSGADT'
    
    ext: Extension = extensions[1]
    assert ext.name == 'NAME'
    assert ext.type == 'ANOTHERFORMAT'
    assert ext.data == 'BLUBBER'

    




from abc import ABC, abstractproperty, abstractstaticmethod

from labfreed.validation import BaseModelWithValidationMessages
from labfreed.PAC_ID.data_model import PACID


    

class Extension(ABC, BaseModelWithValidationMessages): 
    
    @abstractproperty
    def name(self)->str:
        pass
    
    @abstractproperty
    def type(self)->str:
        pass
    
    @abstractproperty
    def data(self)->str:
        pass
    
    @abstractstaticmethod
    def from_spec_fields(*, name, type, data):
        pass
    
    def __str__(self):
        return f'{self.name}${self.type}/{self.data}'
    
    

class UnknownExtension(Extension):
    name_:str
    type_:str
    data_:str
    
    @property
    def name(self)->str:
        return self.name_
    
    @property
    def type(self)->str:
        return self.type_
    
    @property
    def data(self)->str:
        return self.data_
       
    @staticmethod
    def from_spec_fields(*, name, type, data):
        return UnknownExtension(name_=name, type_=type, data_=data)


import re
import pytest

from labfreed.TREX.data_model import *
from labfreed.TREX.parse import _from_trex_string, TREX_Parser

from labfreed.validation import LabFREEDValidationError

parser = TREX_Parser()

    
def test_parse_followed_by_serialization_has_no_effect():
    d = [
        'DATE$T.D:20250101',
        'BOOL$T.B:T',
        'ALPHANUM$T.A:ABC',
        'TXT$T.T:ABCD',
        'BIN$T.X:ABDF',
        'ERR$E:123',
        'DUR$HUR:10',
        'TAB$$C-0$T.A:C.1$T.B::TRUE:T::FALSE:F'        
    ]
    for trex_str in d:
        trex = parser.parse_trex_str(trex_str, name='A')
        assert trex_str == trex.data
        
    trex_str = '+'.join(d)
    trex = parser.parse_trex_str(trex_str, name='A')
    assert trex_str == trex.data


import pytest
from labfreed.PAC_ID.data_model import IDSegment
from labfreed.IO.parse_pac import PAC_Parser


valid_base = "HTTPS://PAC.METTORIUS.COM/"
valid_standard_segments = "-MD/240:B-800/21:12345"
valid_dummy_extension = "DUMMY$MYTYPE/DUMMYDATA"


parser = PAC_Parser()



# Issuer
def test_standard_base_gives_correct_issuer():
    pac = parser.parse("HTTPS://PAC.METTORIUS.COM/" + valid_standard_segments)
    assert pac.pac_id.issuer == "METTORIUS.COM"
       
def test_pac_can_be_missing_from_domain():
    pac = parser.parse("HTTPS://METTORIUS.COM/" + valid_standard_segments)
    assert pac.pac_id.issuer == "METTORIUS.COM"
     
def test_pac_can_be_missing_from_domain():
    pac = parser.parse("METTORIUS.COM/" + valid_standard_segments)
    assert pac.pac_id.issuer == "METTORIUS.COM"
    
def test_issuer_must_be_valid_domain():
    with pytest.raises(Exception):
        pac = parser.parse("HTTPS://METTORIUS/" + valid_standard_segments)
        
        
        
        

# Identifier Segments
def test_pac_must_have_at_least_one_segment():
    with pytest.raises(Exception):
        pac = parser.parse(valid_base = "").pac_id
          
def test_identifier_named_segment():
    pac = parser.parse(valid_base + "KEY:VAL")
    seg: IDSegment = pac.pac_id.identifier[0]
    assert seg.key == "KEY"
    assert seg.value == "VAL"
    
def test_identifier_unnamed_segment():
    pac = parser.parse(valid_base + "VAL")
    seg: IDSegment = pac.pac_id.identifier[0]
    assert not seg.key
    assert seg.value == "VAL"
     
def test_identifier_combination_of_named_and_unnamed_segments():
    pac = parser.parse(valid_base + "KEY0:VAL0/VAL1/KEY2:VAL2")
    pac = pac.pac_id
    seg: IDSegment = pac.identifier[0]
    assert seg.key == 'KEY0'
    assert seg.value == "VAL0"
    
    seg: IDSegment = pac.identifier[1]
    assert not seg.key
    assert seg.value == "VAL1"
    
    seg: IDSegment = pac.identifier[2]
    assert seg.key == 'KEY2'
    assert seg.value == "VAL2"
      
def test_keys_must_be_unique():
    pac = parser.parse(valid_base + "KEY:VAL/KEY:ANOTHERVAL/KEY:VAL/KEY2:ANOTHERVAL")
    assert len(pac.pac_id.get_warnings()) > 0
        
def test_keys_can_repeat_accross_categories():
    pac = parser.parse(valid_base + "-MD/KEY:VAL/-MS/KEY:VAL")
    assert True # made it here without exception > it's fine
        
        

    



import pytest
from labfreed.PAC_CAT.data_model import Category, PAC_CAT
from labfreed.PAC_ID.data_model import IDSegment
from labfreed.PAC_ID.extensions import Extension
from labfreed.IO.parse_pac import PAC_Parser
from labfreed.validation import LabFREEDValidationError


valid_base = "HTTPS://PAC.METTORIUS.COM/"
valid_standard_segments = "-MD/240:B-800/21:12345"
valid_dummy_extension = "DUMMY$MYTYPE/DUMMYDATA"


parser = PAC_Parser()


# Identifier Segments Categories (Recommendation)
def test_basic_valid_category():
    pac = parser.parse(valid_base + "-DM/21:VAL")
    cat: Category = pac.pac_id.categories[0]
    assert cat.key == '-DM'
    assert cat.segments[0].key == '21'
    assert cat.segments[0].value == 'VAL'
    
def test_if_no_category_is_specified_default_to_unnamed_category():
    pac = parser.parse(valid_base + "KEY:VAL")
    cat: Category = pac.pac_id.categories[0]
    assert cat.key == None
    assert cat.segments[0].key == 'KEY'
    assert cat.segments[0].value == 'VAL'
    
def test_missing_dash_in_category_name_is_interpreted_as_unnamed_segment():
    pac = parser.parse(valid_base + "DM/21:VAL")
    cat: Category = pac.pac_id.categories[0]
    assert cat.key == None
    assert cat.segments[0].value == 'DM'
    
def test_category_with_multiple_segments():
    pac = parser.parse(valid_base + "-MX/KEY0:VAL0/VAL1/KEY2:VAL2")
    cat: Category = pac.pac_id.categories[0]
    assert cat.key == '-MX'
    
    seg: IDSegment = cat.segments[0]
    assert seg.key == 'KEY0'
    assert seg.value == "VAL0"
    
    seg: IDSegment = cat.segments[1]
    assert not seg.key
    assert seg.value == "VAL1"
    
    seg: IDSegment = cat.segments[2]
    assert seg.key == 'KEY2'
    assert seg.value == "VAL2"
      
def test_two_categories():
    pac = parser.parse(valid_base + "-DX/KEY:VAL/-MX/KEY:VAL")
    cat: Category = pac.pac_id.categories[0]
    assert cat.key == '-DX'
    assert cat.segments[0].key == 'KEY'
    assert cat.segments[0].value == 'VAL'
    
    cat: Category = pac.pac_id.categories[1]
    assert cat.key == '-MX'
    assert cat.segments[0].key == 'KEY'
    assert cat.segments[0].value == 'VAL'
     
def test_three_categories():
    pac = parser.parse(valid_base + "-DX/KEY0:VAL0/-MX/KEY1:VAL1/-CAT/KEY2:VAL2")
    pac = pac.pac_id
    cat: Category = pac.categories[0]
    assert cat.key == '-DX'
    assert cat.segments[0].key == 'KEY0'
    assert cat.segments[0].value == 'VAL0'
    
    cat: Category = pac.categories[1]
    assert cat.key == '-MX'
    assert cat.segments[0].key == 'KEY1'
    assert cat.segments[0].value == 'VAL1'
    
    cat: Category = pac.categories[2]
    assert cat.key == '-CAT'
    assert cat.segments[0].key == 'KEY2'
    assert cat.segments[0].value == 'VAL2'
    
def test_implied_segments_of_MD_category():
    pac = parser.parse(valid_base + "-MD/0/1")
    cat: Category = pac.pac_id.categories[0]
    assert cat.segments[0].key == '240'
    assert cat.segments[1].key == '21'
    
def test_implied_segments_of_MS_category():
    pac = parser.parse(valid_base + "-MS/0/1/2/3/4")
    cat: Category = pac.pac_id.categories[0]
    assert cat.segments[0].key == '240'
    assert cat.segments[1].key == '10'
    assert cat.segments[2].key == '20'
    assert cat.segments[3].key == '21'
    assert cat.segments[4].key == '250'
    
def test_implied_segments_of_MC_category():
    pac = parser.parse(valid_base + "-MC/0/1/2/3/4")
    cat: Category = pac.pac_id.categories[0]
    assert cat.segments[0].key == '240'
    assert cat.segments[1].key == '10'
    assert cat.segments[2].key == '20'
    assert cat.segments[3].key == '21'
    assert cat.segments[4].key == '250'
    
def test_implied_segments_of_MM_category():
    pac = parser.parse(valid_base + "-MM/0/1/2/3/4")
    cat: Category = pac.pac_id.categories[0]
    assert cat.segments[0].key == '240'
    assert cat.segments[1].key == '10'
    assert cat.segments[2].key == '20'
    assert cat.segments[3].key == '21'
    assert cat.segments[4].key == '250'
    
def test_stop_implying_segments_after_an_explicit_one_is_found():
    pac = parser.parse(valid_base + "-MS/0/1/KEY:2/3/4")
    cat: Category = pac.pac_id.categories[0]
    assert cat.segments[0].key == '240'
    assert cat.segments[1].key == '10'
    assert cat.segments[2].key == 'KEY'
    assert cat.segments[3].key == None
    assert cat.segments[4].key == None
    
def test_more_segments_than_implicit_keys():
    pac = parser.parse(valid_base + "-MD/IMPLIED1/IMPLIED2/ADDITIONAL")
    cat: Category = pac.pac_id.categories[0]
    assert cat.segments[2].key == None
    
    
    
    
    
def test_mandatory_fields_of_MD_category():
    with pytest.raises(LabFREEDValidationError):
        pac = parser.parse(valid_base + "-MD/KEY:VAL")
    
    pac:PAC_CAT = parser.parse(valid_base + "-MD/21:1234/KEY:VAL/240:BAL500").pac_id
    assert pac.get_category('-MD').serial_number == '1234'
    assert pac.get_category('-MD').model_number == 'BAL500'
    

    



from concurrent.futures import ThreadPoolExecutor, as_completed
from enum import Enum, auto
from pydantic import BaseModel, Field, field_validator
from requests import RequestException, request, ConnectionError, ReadTimeout
from rich import print
from rich.table import Table

from labfreed.validation import BaseModelWithValidationMessages


class CITEntry(BaseModelWithValidationMessages):
    service_name: str
    application_intents:list[str]
    service_type:str
    template_url:str
    

class CITBlock(BaseModelWithValidationMessages):
    applicable_if: str  = Field(default='True', alias='if')
    entries: list[CITEntry]
    
    @field_validator('applicable_if', mode='before')
    @classmethod
    def convert_if(cls, v):
        return v if v is not None else 'True'
    

class CIT(BaseModelWithValidationMessages):
    origin: str = ''
    macros:dict = Field(default_factory=dict)
    cit: list[CITBlock] = Field(default_factory=list)
    
    

class ServiceState(Enum):
    ACTIVE = auto()
    INACTIVE = auto()
    UNKNOWN = auto()
    
class Service(BaseModelWithValidationMessages):
    service_name: str
    application_intents:list[str]
    service_type:str
    url:str
    active:ServiceState =ServiceState.UNKNOWN
    
    def check_state(self):
        try:
            r = request('get',self.url, timeout=2)
            if r.status_code < 400:
                self.active = ServiceState.ACTIVE
            else: 
                self.active = ServiceState.INACTIVE
        except RequestException as e:
            print(f"Request failed: {e}")
            self.active = ServiceState.INACTIVE
            
    
class CITEvaluated(BaseModelWithValidationMessages):
    origin: str = ""
    services: list[Service] = Field(default_factory=list)
    
    def update_states(self):
        with ThreadPoolExecutor(max_workers=10) as executor:
            futures = [executor.submit(s.check_state) for s in self.services]
            for _ in as_completed(futures):
                pass  # just wait for all to finish
    
    def __str__(self):
        out = [f'CIT (origin {self.origin})']
        for s in self.services:
            out.append(f'{s.service_name}\t\t\t{s.url}')
        return '\n'.join(out)
    
    def print(self):
        table = Table(title=f"Services from origin '{self.origin}")

        table.add_column("Service Name")
        table.add_column("URL")
        table.add_column('Reachable')
        
        for s in self.services:
            table.add_row(s.service_name, s.url, s.active.name)

        print(table)


import re
import pytest

from labfreed.TREX.data_model import *
from labfreed.TREX.parse import _from_trex_string, TREX_Parser

from labfreed.validation import LabFREEDValidationError

parser = TREX_Parser()

    
def test_trex_parse():
    trex_str = 'SUM$TREX/A$T.A:ASDFAS+B$T.B:T'
    trex = parser.parse_trex_str(trex_str)
    assert trex.name == 'SUM'
    assert trex.type == 'TREX'
    seg = trex.segments
    assert trex.get_segment('A').type == 'T.A'
    assert trex.get_segment('A').value == 'ASDFAS'
    assert trex.get_segment('B').type == 'T.B'
    assert trex.get_segment('B').value == 'T'
    
    
    
    
# Segment Key   
def test_invalid_keys():
    trex_str = '£TIM$HUR:1'
    with pytest.raises(LabFREEDValidationError) as e:
        trex = parser.parse_trex_str(trex_str, name="A")
        
    trex_str = 'tImE$HUR:1'
    with pytest.raises(LabFREEDValidationError) as e:
        trex = parser.parse_trex_str(trex_str, name="A")


# Numeric Segment
def test_numeric_segment_valid():
    valid_strs = [
        'TIME$HUR:25.0',
        'TIME$HUR:25',
        'TIME$HUR:-25.0',
        'TIME$HUR:-25'
    ]
    for trex_str in valid_strs:
        trex = parser.parse_trex_str(trex_str, name="A")
        assert trex.data == trex_str
    

        
def test_numeric_segment_scientific_valid():   
    valid_strs = [
        'TIME$HUR:-25E2',
        'TIME$HUR:-25E-2',
        'TIME$HUR:-25.12E2',
        'TIME$HUR:-25.12E-2'
    ]
    for trex_str in valid_strs:
        trex = parser.parse_trex_str(trex_str, name="A")
        assert trex.data == trex_str

    
    
def test_numeric_segment_invalid_values():
    with pytest.raises(LabFREEDValidationError):
        trex_str ='TIME$HUR:A'
        trex = parser.parse_trex_str(trex_str, name="A")
        assert trex.data == trex_str
    
    with pytest.raises(LabFREEDValidationError):
        trex_str ='TIME$HUR:1.1.1'
        trex = parser.parse_trex_str(trex_str, name="A")
        assert trex.data == trex_str
        
def test_numeric_segment_invalid_unit():
    with pytest.raises(LabFREEDValidationError):
        trex_str ='TIME$HIP:A'
        trex = parser.parse_trex_str(trex_str, name="A")
        assert trex.data == trex_str
    
        
        
# DateSegments
def test_valid_dates():
    dates = [
        '20240305'
    ]
    for e in dates:
        trex = parser.parse_trex_str(f'A$T.D:{e}', name='A')
        assert not trex.get_nested_validation_messages()

def test_valid_times():
    times = [
        'T0102', 'T010203', 'T010203.456'
    ]
    for e in times:
        trex = parser.parse_trex_str(f'A$T.D:{e}', name='A')
        assert not trex.get_nested_validation_messages()
    
def test_valid_datetimes():
    datetimes = [
        '20240305T0102', '20240305T010203', '20240305T010203', '20240305T010203.456'
    ]
    for e in datetimes:
        trex = parser.parse_trex_str(f'A$T.D:{e}', name='A')
        assert not trex.get_nested_validation_messages()
        
def test_invalid_dates_format():
    dates = [
        '2024030'
    ]
    for e in dates:
        with pytest.raises(LabFREEDValidationError):
            trex = parser.parse_trex_str(f'A$T.D:{e}', name='A')
            
            
def test_invalid_dates():
    dates = [
        '20241331', '20240000'
    ]
    for e in dates:
        with pytest.raises(LabFREEDValidationError):
            trex = parser.parse_trex_str(f'A$T.D:{e}', name='A')
            
            
def test_invalid_times_format():
    times = [
        'T2561'
    ]
    for e in times:
        with pytest.raises(LabFREEDValidationError):
            trex = parser.parse_trex_str(f'A$T.D:{e}', name='A')
    


# BoolSgments
def test_valid_bool():
    b = [
        'T', 'F'
    ]
    for e in b:
        trex = parser.parse_trex_str(f'A$T.B:{e}', name='A')
        assert not trex.get_nested_validation_messages()
        
def test_invalid_bool():
    b = [
        '1', '0', 'TRUE', 'FALSE'
    ]
    for e in b:
        with pytest.raises(LabFREEDValidationError):
            trex = parser.parse_trex_str(f'A$T.B:{e}', name='A')
            
            
            
# Alphanumeric Segment
def test_valid_alphanumeric():
    b = [
        'ABCDEFGHIJKLMNOPQRSTUVW012345678.-', 
    ]
    for e in b:
        trex = parser.parse_trex_str(f'A$T.A:{e}', name='A')
        assert not trex.get_nested_validation_messages()
        
def test_invalid_alphanumeric():
    b = [
        'a', '£', '<', 'ABCDeFGh'
    ]
    for e in b:
        with pytest.raises(LabFREEDValidationError):
            trex = parser.parse_trex_str(f'A$T.A:{e}', name='A')



# Text Segment
def test_valid_text():
    b = [
        'ABCDEFGHIJKLMNOPQRSTUVW012345678', 
        'ABCD'
    ]
    for e in b:
        trex = parser.parse_trex_str(f'A$T.T:{e}', name='A')
        assert not trex.get_nested_validation_messages()
        assert trex.get_segment('A').value == e
          
def test_invalid_text():
    b = [
        'a', '£', '<', 'ABCDeFGh', 'ABCDEFGHIJKLMNOPQRSTUVW012345678.-'
    ]
    for e in b:
        with pytest.raises(ValueError):
            trex = parser.parse_trex_str(f'A$T.X:{e}', name='A')
            
def test_valid__with_b36_conversion_text():
    b = [
        ('🐯', '1URIOQ7')
    ]
    for e in b:
        trex = TREX(segments=[TextSegment(key='A', value=e[0])], name_='A')
        assert not trex.get_nested_validation_messages()
        assert trex.get_segment('A').value == e[1]
        

      
# Binary Segment
def test_valid_binary():
    b = [
        'ABCDEFGHIJKLMNOPQRSTUVW012345678'
    ]
    for e in b:
        trex = parser.parse_trex_str(f'A$T.X:{e}', name='A')
        assert not trex.get_nested_validation_messages()
        
def test_invalid_binary():
    b = [
        'a', '£', '<', 'ABCDeFGh', 'ABCDEFGHIJKLMNOPQRSTUVW012345678.-'
    ]
    for e in b:
        with pytest.raises(ValueError):
            trex = parser.parse_trex_str(f'A$T.X:{e}', name='A')
            
            
# Error Segment
def test_valid_error():
    b = [
        'ABCDEFGHIJKLMNOPQRSTUVW012345678.-'
    ]
    for e in b:
        trex = parser.parse_trex_str(f'A$E:{e}', name='A')
        assert not trex.get_nested_validation_messages()
        
def test_invalid_error():
    b = [
        'a', '£', '<', 'ABCDeFGh'
    ]
    for e in b:
        with pytest.raises(LabFREEDValidationError):
            trex = parser.parse_trex_str(f'A$E:{e}', name='A')
            
            
# Table
def test_valid_table():
    tab = 'TAB$$C-0$T.A:C.1$T.B::TRUE:T::FALSE:F'
    trex = parser.parse_trex_str(tab, name='A')
    assert not trex.get_errors()
    t = trex.segments[0]
    assert isinstance(t, TREX_Table)
    assert t.column_headers[0].key == 'C-0'
    assert t.column_headers[0].type == 'T.A'
    assert t.column_headers[1].key == 'C.1'
    assert t.column_headers[1].type == 'T.B'
    assert len(t.data) == 2
    assert [e.value for e in t.data[0]] == ['TRUE', 'T']
    assert [e.value for e in t.data[1]] == ['FALSE', 'F']
    
def test_invalid_header_keys():
    tab = 'TAB$$C£0$T.A:C1$T.B::TRUE:T::FALSE:F' # character £ in key
    with pytest.raises(LabFREEDValidationError):
        trex = parser.parse_trex_str(tab, name='A')
        
def test_valid_header_types():
    d = [
        ('T.D', '20240101'), 
        ('T.B', 'T'), 
        ('T.A', 'ABC'),
        ('T.T', 'ABC'), 
        ('T.X', 'ABC'),
        ('E', 'ABC'), 
        ('HUR', '1'),
        ('C63', '2')
        ]
    for e in d:
        tab = f'TAB$$A${e[0]}::{e[1]}'
        trex = parser.parse_trex_str(tab, name='A')
        assert not trex.get_errors()
        
def test_invalid_header_types():
    types = ['T.Q', 'T.', 'HURR', 'W70']
    for t in types:
        tab = f'TAB$$A${t}::V'
        with pytest.raises(LabFREEDValidationError):
            trex = parser.parse_trex_str(tab, name='A')
            
def test_size_mismatch():
    tab = 'TAB$$C0$T.A:C1$T.B:C2$C63::TRUE:T::FALSE:T:1' #row 0 has only 2 elements
    with pytest.raises(LabFREEDValidationError):
        trex = parser.parse_trex_str(tab, name='A')
        
def test_type_mismatch():
    tab = 'TAB$$C0$T.A:C1$T.B::TRUE:TRUE::FALSE:T' #element (0,1) has wrong type. should be boolean (T or F) but is text 
    with pytest.raises(LabFREEDValidationError):
        trex = parser.parse_trex_str(tab, name='A')
        
            


import os
import re
import io
import contextlib

start_marker = "<!-- BEGIN EXAMPLES -->"
end_marker = "<!-- END EXAMPLES -->"

with open("examples.py", encoding="utf-8") as ex_file:
    example_content = ex_file.read()

# Match blocks: either triple-quoted docstrings or plain code chunks
pattern = r"(?P<docstring>'''(.*?)'''|\"\"\"(.*?)\"\"\")"
matches = list(re.finditer(pattern, example_content, flags=re.DOTALL))

output_parts = []
last_index = 0

shared_globals = {
    "__file__": os.path.abspath("examples.py")
}

def run_code_block(code_str):
    stdout = io.StringIO()
    try:
        with contextlib.redirect_stdout(stdout):
            exec(code_str, shared_globals)
    except Exception as e:
        return f"[Error during execution: {e}]"
    return stdout.getvalue().strip()


def format_output_as_html(output: str) -> str:
    lines = output.splitlines()
    formatted = "\n".join(f">> {line}" for line in lines)
    html_output = f"<pre><code style=\"color: green\">{formatted}</code></pre>"
    return html_output


def format_output(output: str) -> str:
    lines = output.splitlines()
    formatted = "\n".join(f">> {line}" for line in lines)
    return  f"```text\n{formatted}\n```"


for match in matches:
    # Capture code before this docstring
    code_chunk = example_content[last_index:match.start()]
    if code_chunk.strip():
        code_block = code_chunk.strip("\n")
        output_parts.append(f"```python\n{code_block}\n```")

        result = run_code_block(code_block)
        if result:
            output_parts.append(format_output(result))

    # Capture the docstring itself
    doc = match.group(0).strip('\'" \n')
    if doc:
        output_parts.append(doc.strip() + "\n")

    last_index = match.end()

# Add any remaining code after the last docstring
if last_index < len(example_content):
    remaining_code = example_content[last_index:].strip("\n")
    if remaining_code.strip():
        output_parts.append(f"```python\n{remaining_code}\n```")

        result = run_code_block(remaining_code)
        if result:
            output_parts.append(format_output(result))

# Combine the parts
example_content_md = "\n".join(part for part in output_parts if part.strip())

# Inject into README.md
with open("README.md", encoding="utf-8") as readme_file:
    readme = readme_file.read()

if not readme:
    raise ValueError("README.md is empty")

if start_marker not in readme or end_marker not in readme:
    raise ValueError("Markers not found in README.md")

new_readme = (
    readme.split(start_marker)[0]
    + start_marker + "\n"
    + example_content_md + "\n"
    + end_marker
    + readme.split(end_marker)[1]
)

with open("README.md", "w", encoding="utf-8") as readme_file:
    readme_file.write(new_readme)



import logging
from pydantic import BaseModel
from labfreed.PAC_ID.extensions import Extension
from labfreed.utilities.base36 import from_base36, to_base36


class DisplayNames(Extension, BaseModel):
    display_names: list[str]
       
    @property
    def name(self)->str:
        return 'N'
    
    @property
    def type(self)->str:
        return 'N'
    
    @property
    def data(self)->str:
        return '/'.join([to_base36(dn) for dn in self.display_names])
    
    @staticmethod
    def from_spec_fields(*, name, type, data):
        if name != 'N':
            logging.warning(f'Name {name} was given, but this extension should only be used with name "N". Will ignore input')
            
        if type != 'N':
            logging.warning(f'Type {name} was given, but this extension should only be used with type "N". Will try to parse data as display names')
        
        display_names = [from_base36(b36) for b36 in data.split('/')]
         
        return DisplayNames(display_names=display_names)
    
    def __str__(self):
        return 'Display names: '+ ';'.join(self.display_names)





import os
import re
import yaml
import json
import jsonpath_ng.ext as jsonpath


from labfreed.IO.parse_pac import PAC_Parser, PACID_With_Extensions
from labfreed.PAC_ID_Resolver.data_types import CIT, CITEntry, CITEvaluated, Service

from labfreed.PAC_ID_Resolver.non_needed.query_tools import JSONPathTools


def load_cit(path):
    with open(path, 'r') as f:
        cit = yaml.safe_load(f)
        cit = CIT.model_validate(cit)
        return cit


class PAC_ID_Resolver():
    def __init__(self, cits:list[CIT]=None):
        if not cits:
            cits = []
        self.cits = cits
        
        # load the default cit
        dir = os.path.dirname(__file__)
        fn ='cit.yaml'
        p = os.path.join(dir, fn)       
        with open(p, 'r') as f:
            cit = yaml.safe_load(f)
            cit = CIT.model_validate(cit)
            self.cits.append(cit)
        
        
    def resolve(self, pac_id:PACID_With_Extensions|str):
        if isinstance(pac_id, str):
            pac_id = PAC_Parser().parse(pac_id)
        
        pac_id_json = pac_id.model_dump(by_alias=True)
        
        
        # dir = os.path.dirname(__file__)
        # p = os.path.join(dir, 'pac-id.json')
        # with open(p , 'r') as f:
        #     _json = f.read()
        #     pac_id_json = json.loads(_json)
        
        matches = [self._evaluate_against_cit(pac_id_json, cit) for cit in self.cits]
        return matches
            

    def _evaluate_against_cit(self, pac_id_json, cit:CIT):
        cit_evaluated = CITEvaluated(origin=cit.origin)   
        for block in cit.cit:
            _, is_applicable = self._evaluate_applicable_if(pac_id_json, block.applicable_if)
            if not is_applicable:
                continue

            for e in block.entries:
                url = self.eval_url_template(pac_id_json, e.template_url)
                cit_evaluated.services.append(Service(  
                                                        service_name=e.service_name,
                                                        application_intents=e.application_intents,
                                                        service_type=e.service_type,
                                                        url = url
                                    )
                              )            
        return cit_evaluated
    
    
    def _evaluate_applicable_if(self, pac_id_json:str, expression) -> tuple[str, bool]:
        expression = self._apply_convenience_substitutions(expression)
        
        tokens = self._tokenize_jsonpath_expression(expression)
        expression_for_eval = self._expression_from_tokens(pac_id_json, tokens)
        applicable = eval(expression_for_eval, {}, {})
        
        return expression_for_eval, applicable
    
    
    def _apply_convenience_substitutions(self, query):
        ''' applies a few substitutions, which enable abbreviated syntax.'''
        
        # allow access to array elements by key
        q_mod = re.sub(r"\[('.+?')\]", r"[?(@.key == \1)]", query )
        
        # allow shorter path
        # substitutions = [
        #     (r'(?<=^)id', 'pac.id'),
        #     (r'(?<=^)cat', 'pac.id.cat'),
        #     (r'(?<=\.)id(?=\.)', 'identifier'),
        #     (r'(?<=\.)cat$', 'categories'),
        #     (r'(?<=\.)cat(?=\[)', 'categories'),
        #     (r'(?<=\.)seg$', 'segments'),
        #     (r'(?<=\.)seg(?=\[)', 'segments'),
        #     (r'(?<=^)isu', 'pac.isu'),
        #     (r'(?<=\.)isu', 'issuer'),
        #     (r'(?<=^)ext', 'pac.ext'),
        #     (r'(?<=\.)ext(?=$)', 'extensions'),
        #     (r'(?<=\.)ext(?=\[)', 'extensions'),
        # ]
        # for sub in substitutions:
        #     q_mod = re.sub(sub[0], sub[1], q_mod)
        
        return q_mod
    

    def _tokenize_jsonpath_expression(self, expr: str):
        token_pattern = re.compile(
            r"""
            (?P<LPAREN>\() |
            (?P<RPAREN>\)) |
            (?P<LOGIC>\bAND\b|\bOR\b|\bNOT\b) |
            (?P<OPERATOR>==|!=|<=|>=|<|>) |
            (?P<JSONPATH>
                \$                               # starts with $
                (?:
                    [^\s\[\]()]+                # path segments, dots, etc.
                    |
                    \[                           # open bracket
                        (?:                     # non-capturing group
                            [^\[\]]+            # anything but brackets
                            |
                            \[[^\[\]]*\]        # nested brackets (1 level)
                        )*
                    \]
                )+                              # one or more bracket/segment blocks
            ) |
            (?P<LITERAL>
                [A-Za-z_][\w\.\-]*[A-Za-z0-9]   # domain-like literals
            )
            """,
            re.VERBOSE
        )

        tokens = []
        pos = 0
        while pos < len(expr):
            match = token_pattern.match(expr, pos)
            if match:
                group_type = match.lastgroup
                value = match.group().strip()
                tokens.append((value, group_type))
                pos = match.end()
            elif expr[pos].isspace():
                pos += 1  # skip whitespace
            else:
                raise SyntaxError(f"Unexpected character at position {pos}: {expr[pos]}")

        return tokens
    
    
    def _expression_from_tokens(self, pac_id_json:str, tokens: tuple[str, str]):
        out  = []
        for i in range(len(tokens)):
            prev_token = tokens[i-1] if i > 0 else (None, None)
            curr_token = tokens[i]
            next_token = tokens[i+1] if i < len(tokens)-1 else (None, None)
            if curr_token[1] == 'JSONPATH':
                res = self._evaluate_jsonpath(pac_id_json, curr_token[0])
                
                if prev_token[1] == 'OPERATOR' or next_token[1] == 'OPERATOR':
                    # if token is part of comparison return the value of the node
                    if len(res) == 0:
                        out.append('""')
                    else:
                        out.append(f'"{res[0].upper()}"')
                else:
                    # if token is not part of comparison evaluate to boolean
                    if len(res) == 0:
                        out.append(False)
                    else:
                        out.append(True)
                        
            elif curr_token[1] == 'LOGIC':
                out.append(curr_token[0].lower())  
                
            elif curr_token[1] == 'LITERAL':
                t = curr_token[0]
                if t[0] != '"':
                    t = '"' + t
                if t[-1] != '"':
                    t = t + '"' 
                out.append(t.upper())    
            else:
                out.append(curr_token[0])
                
        s = ' '.join([str(e) for e in out])
        return s
                    

    
    
    def eval_url_template(self, pac_id_json, url_template):
        url = url_template
        placeholders = re.findall(r'\{(.+?)\}', url_template)
        for placeholder in placeholders:
            expanded_placeholder = self._apply_convenience_substitutions(placeholder)
            res = self._evaluate_jsonpath(pac_id_json, expanded_placeholder) or ['']
            url = url.replace(f'{{{placeholder}}}', str(res[0]))
            # res = self.substitute_jsonpath_expressions(expanded_placeholder, Patterns.jsonpath.value, as_bool=False)
            # url = url.replace(f'{{{placeholder}}}', res)
        return url
    

    
    
    def _evaluate_jsonpath(self, pac_id_json, jp_query):
        jsonpath_expr = jsonpath.parse(jp_query)
        matches = [match.value for match in jsonpath_expr.find(pac_id_json)]
        return matches
    
    

    
    
if __name__ == '__main__':
    r = PAC_ID_Resolver()
    r.resolve()





import re
from types import MappingProxyType

from labfreed.DisplayNameExtension.DisplayNameExtension import DisplayNames
from labfreed.PAC_CAT.data_model import PAC_CAT
from labfreed.PAC_ID.extensions import Extension, UnknownExtension
from labfreed.TREX.data_model import TREX


from ..PAC_ID.data_model import *

from ..validation import ValidationMessage, LabFREEDValidationError






class PACID_With_Extensions(BaseModelWithValidationMessages):
    pac_id: PACID = Field(serialization_alias='pac')
    extensions: list[Extension] = Field(default_factory=list)
    
    def __str__(self):
        out = str(self.pac_id)
        out += '*'.join(str(e) for e in self.extensions)
        return out
        
    def get_extension_of_type(self, type:str) -> list[Extension]:
        return [e for e in self.extensions if e.type == type]
    
    def get_extension(self, name:str) -> Extension|None:
        out = [e for e in self.extensions if e.name == name]
        if not out:
            return None
        return out[0]
    
    
    def serialize(self, use_short_notation_for_extensions=False, uppercase_only=False):
        extensions_str = self._serialize_extensions(self.extensions, use_short_notation_for_extensions)
        out = self.pac_id.serialize() + extensions_str
        if uppercase_only:
            out = out.upper()
        return out
        
    def to_url(self, use_short_notation_for_extensions=False, uppercase_only=False) -> str:
        return self.serialize(use_short_notation_for_extensions, uppercase_only)
    
    @classmethod
    def deserialize(cls, url, extension_interpreters ):
        parser = PAC_Parser(extension_interpreters)
        return parser.parse(url)
        
        
    

    def _serialize_extensions(self, extensions:list[Extension], use_short_notation_for_extensions):
        out = ''
        short_notation = use_short_notation_for_extensions
        for i, e in enumerate(extensions):
            
            if short_notation and i==0:
                if e.name=='N':
                    out += f'*{e.data}'
                    continue
                else: 
                    short_notation = False
            if short_notation and i==1:
                if e.name=='SUM':
                    out += f'*{e.data}'
                    continue
                else: 
                    short_notation = False
                
            out += f'*{e.name}${e.type}/{e.data}'
        return out
        







class PAC_Parser():
    
    def __init__(self, extension_interpreters:dict[str, Extension]=None):
        self.extension_interpreters = extension_interpreters or {'TREX': TREX, 'N': DisplayNames}
        
    def parse(self, pac_url:str, suppress_errors=False) -> PACID_With_Extensions:
        if '*' in pac_url:
            id_str, ext_str = pac_url.split('*', 1)
        else:
            id_str = pac_url
            ext_str = ""
            
        pac_id = self._parse_pac_id(id_str)
        extensions = self._parse_extensions(ext_str)
        
        pac_with_extension = PACID_With_Extensions(pac_id=pac_id, extensions=extensions)
        if not pac_with_extension.is_valid() and not suppress_errors:
            raise LabFREEDValidationError(validation_msgs = pac_with_extension.get_nested_validation_messages(), model=pac_with_extension)
        
        return pac_with_extension
            
            
    def _parse_pac_id(self,id_str:str) -> PACID:
        m = re.match(f'(HTTPS://)?(PAC.)?(?P<issuer>.+?\..+?)/(?P<identifier>.*)', id_str)
        d = m.groupdict()
        
        id_segments = list()
        default_keys = None
        id_segments = self._parse_id_segments(d.get('identifier'))
        
        pac = PACID(issuer= d.get('issuer'),
                     identifier=id_segments
        )
        
        # if a segment starts with '-' the pac is interpreted as category
        if any([s for s in pac.identifier if '-' in s.value]):
            pac = PAC_CAT.from_pac_id(pac)
        
        return pac
        

    
    
    def _parse_id_segments(self, identifier:str):
        if not identifier:
            return []    
        
        id_segments = list()  
        if len(identifier) > 0 and identifier[0] == '/':
            identifier = identifier[1:]
        for s in identifier.split('/'):
            tmp = s.split(':')
            
            if len(tmp) == 1:
                segment = IDSegment(value=tmp[0])
            elif len(tmp) == 2:
                segment = IDSegment(key=tmp[0], value=tmp[1])
            else:
                raise ValueError(f'invalid segment: {s}')
                
            id_segments.append(segment)
        return id_segments
    
    


    def _parse_extensions(self, extensions_str:str|None) -> list[Extension]:    
        
        extensions = list()
        
        if not extensions_str:
            return extensions
        
        defaults =  MappingProxyType(
                                {
                                    0: { 'name': 'N', 'type': 'N'},
                                    1: { 'name': 'SUM', 'type': 'TREX'}
                                }
        )
        for i, e in enumerate(extensions_str.split('*')):
            if e == '': #this will happen if first extension starts with *
                continue
            d = re.match('((?P<name>.+)\$(?P<type>.+)/)?(?P<data>.+)', e).groupdict()
            
            name = d.get('name')
            type = d.get('type') 
            data = d.get('data')
    
            if name:
                defaults = None # once a name was specified no longer assign defaults
            else:
                if defaults:
                    name = defaults.get(i).get('name')
                    type = defaults.get(i).get('type')
                else:
                    raise ValueError('extension number {i}, must have name and type')
            
            #convert to subtype if they were given
            subtype = self.extension_interpreters.get(type) or UnknownExtension
            e = subtype.from_spec_fields(name=name, type=type, data=data)
            extensions.append(e)

        return extensions
        
 


from labfreed.PAC_ID.data_model import PACID, IDSegment

pac_id = PACID(issuer = 'METTORIUS.COM',
            identifier = [  IDSegment(value='-DR'), 
                            IDSegment(value='999'),
                            IDSegment(value="-MD"),
                            IDSegment(key='240', value='1'),
                            IDSegment(key='21', value='2')
                        ]
          ) 

 
def test_url_serialization():
    url = pac_id.serialize()
    assert url == 'HTTPS://PAC.METTORIUS.COM/-DR/999/-MD/240:1/21:2'


import pytest
from pydantic import ValidationError

from labfreed.PAC_CAT.data_model import *

additional_segments=[IDSegment(key='K1', value='V1'),
                                         IDSegment(value='V2')]



def test_alias():
    md = Material_Device(serial_number='123', model_number='aa')
    md2 = Material_Device(**{'21': '123', '240': 'aa'})
    assert md == md2


md = Material_Device(model_number='BAL500', 
                    serial_number='1234', 
                    additional_segments=[IDSegment(key='K1', value='V1'),
                                         IDSegment(value='V2')])
def test_required_fields_MD():
    with pytest.raises(ValidationError):
        Material_Device(serial_number=None, model_number=None)
        

def test_MD_segments():
    cat = md.to_identifier_category()
    assert cat.key == '-MD'
    assert cat.segments[0].key == '240'
    assert cat.segments[0].value == 'BAL500'
    assert cat.segments[1].key == '21'
    assert cat.segments[1].value == '1234'
    

    
    
def test_additional_segments():
    cat = md.to_identifier_category()
    assert cat.segments[2].key == 'K1'
    assert cat.segments[2].value == 'V1'
    assert cat.segments[3].key is None
    assert cat.segments[3].value == 'V2'
    
    
ms = Material_Substance(product_number='X67678',
                        batch_number='9999',
                        container_size='1000',
                        container_number='34',
                        aliquot='2',
                        additional_segments=additional_segments
                        )



def test_MS_segments():
    cat = ms.to_identifier_category()
    assert cat.key == '-MS'
    assert cat.segments[0].key == '240'
    assert cat.segments[0].value == 'X67678'
    assert cat.segments[1].key == '10'
    assert cat.segments[1].value == '9999'
    assert cat.segments[2].key == '20'
    assert cat.segments[2].value == '1000'
    assert cat.segments[3].key == '21'
    assert cat.segments[3].value == '34'
    assert cat.segments[4].key == '250'
    assert cat.segments[4].value == '2'
    
ms_some_empty = Material_Substance(product_number='X67678',
                        batch_number='9999',
                        aliquot='2',
                        additional_segments=additional_segments
                        )
    
def test_empty_segments_not_written():
    cat = ms_some_empty.to_identifier_category()
    assert cat.key == '-MS'
    assert cat.segments[0].key == '240'
    assert cat.segments[0].value == 'X67678'
    assert cat.segments[1].key == '10'
    assert cat.segments[1].value == '9999'
    assert cat.segments[2].key == '250'
    assert cat.segments[2].value == '2'


def test_short_notation():
    cat = ms.to_identifier_category(use_short_notation=True)
    assert cat.key == '-MS'
    assert cat.segments[0].key is None
    assert cat.segments[0].value == 'X67678'
    assert cat.segments[1].key is None
    assert cat.segments[1].value == '9999'
    assert cat.segments[2].key is None
    assert cat.segments[2].value == '1000'
    assert cat.segments[3].key is None
    assert cat.segments[3].value == '34'
    assert cat.segments[4].key is None
    assert cat.segments[4].value == '2'
    
def test_short_notation_break():
    cat = ms_some_empty.to_identifier_category(use_short_notation=True)
    assert cat.key == '-MS'
    assert cat.segments[0].key is None
    assert cat.segments[0].value == 'X67678'
    assert cat.segments[1].key is None
    assert cat.segments[1].value == '9999'
    assert cat.segments[2].key == '250'
    assert cat.segments[2].value == '2'
    
    
    



from functools import cache
import json
from pathlib import Path


@cache
def unece_units() -> list[dict]:
    p = Path(__file__).parent / 'UneceUnits.json'
    with open(p) as f:
        l = json.load(f)
    return l

@cache
def unece_unit_codes():
    codes= [u.get('commonCode') for u in unece_units() if u.get('state') == 'ACTIVE']
    return codes


def unece_unit(unit_code):
    unit =  [u for u in unece_units() if u['commonCode'] == unit_code]
    if len(unit) == 0:
        return None
    else:
        return unit[0]
    
def unit_symbol(unit:dict) ->str:
    return unit.get('symbol')

def unit_name(unit:dict) ->str:
    return unit.get('name')





# def quantity_from_UN_CEFACT(value:str, unit_UN_CEFACT) -> UnitQuantity:
#     """ 
#     Maps units from https://unece.org/trade/documents/revision-17-annexes-i-iii
#     to an object of the quantities library https://python-quantities.readthedocs.io/en/latest/index.html 
#     """
#     # cast to numeric type. try int first, which will fail if string has no decimals.
#     # nothing to worry yet: try floast next. if that fails the input was not a str representation of a number
#     try:
#         value_out = int(value)
#     except ValueError:
#         try: 
#             value_out = float(value)
#         except ValueError as e:
#             raise Exception(f'Input {value} is not a str representation of a number') from e
        
#     d = {um[0]: um[1] for um in unit_map}
    
#     unit = d.get(unit_UN_CEFACT)
#     if not unit:
#         raise NotImplementedError(f"lookup for unit {unit} not implemented")
#     out = UnitQuantity(data=value_out, unit_name=unit.name, unit_symbol=unit.symbol)

#     return out



# def quantity_to_UN_CEFACT(value:UnitQuantity ) -> Tuple[int|float, str]:    
#     d = {um[1].symbol: um[0] for um in unit_map}
    
#     unit_un_cefact = d.get(value.unit_symbol)
#     if not unit_un_cefact:
#         raise NotImplementedError(f"lookup for unit {value.unit_symbol} not implemented")
#     return value.data, unit_un_cefact
    
    



def check_compatibility_unece_quantities():
    unece = get_unece_units()
    print(f'Number of units in file: {len(unece)}')
    
    failed = list()
    sucess = list()
    for u in unece:
        if u.get('state') ==  'ACTIVE':
            try:
                if not u.get('symbol'):
                    assert False
                u.get('name')  
                validate_unit(u.get('symbol'))
                sucess.append(u)
            except AssertionError as e:
                failed.append(u)
        else:
            pass
        
    
    
    print('[blue] FAILED [/blue]')
    for u in failed:
        print(f'{u.get('commonCode')}: {u.get('name')}')
    
    print('[yellow] SUCCESSFUL [/yellow]')
    for u in sucess:
        print(u)
        
    print(f'{len(failed)} / {len(unece)} failed to convert')
        

        


from enum import Enum, auto
import re
from pydantic import BaseModel, Field, PrivateAttr
from typing import List, Set, Tuple

from rich import print
from rich.text import Text
from rich.table import Table


domain_name_pattern = r"(?!-)([A-Za-z0-9-]{1,63}(?<!-)\.)+[A-Za-z]{2,63}"
hsegment_pattern = r"[A-Za-z0-9_\-\.~!$&'()+,:;=@]|%[0-9A-Fa-f]{2}"


class ValidationMsgLevel(Enum):
    ERROR = auto()
    ERROR_AUTO_FIX = auto()
    WARNING = auto()
    RECOMMENDATION = auto()
    INFO = auto()

class ValidationMessage(BaseModel):
    source_id:int
    source:str 
    level: ValidationMsgLevel
    problem_msg:str
    recommendation_msg: str = ""
    highlight:str = "" #this can be used to highlight problematic parts
    highlight_sub_patterns:list[str] = Field(default_factory=list)


    
        
        
    # @property
    # def emphazised_highlight(self):
    #     fmt = lambda s: f'[emph]{s}[/emph]'
        
    #     if not self.highlight_sub_patterns:
    #         return fmt(self.highlight)
          
    #     result = []
    #     for c in self.highlight:
    #         if c in self.highlight_sub_patterns:
    #             result.append(fmt(c))
    #         else:
    #             result.append(c)

    #     return ''.join(result)

    
class LabFREEDValidationError(ValueError):
    def __init__(self, message=None, validation_msgs=None):
        super().__init__(message)
        self._validation_msgs = validation_msgs

    @property
    def validation_msgs(self):
        return self._validation_msgs
    
    


class BaseModelWithValidationMessages(BaseModel):
    """ Extension of Pydantic BaseModel, so that validator can issue warnings.
    The purpose of that is to allow only minimal validation but on top check for stricter recommendations"""
    _validation_messages: list[ValidationMessage] = PrivateAttr(default_factory=list)

    def add_validation_message(self, *, msg: str, level:ValidationMsgLevel, recommendation:str="", source:str="", highlight_pattern="", highlight_sub=None):
        if not highlight_sub:
            highlight_sub = []
        w = ValidationMessage(problem_msg=msg, recommendation_msg=recommendation, source=source, level=level, highlight=highlight_pattern, highlight_sub_patterns=highlight_sub, source_id=id(self))

        if not w in self._validation_messages:
            self._validation_messages.append(w)

    def get_validation_messages(self) -> list[ValidationMessage]:
        return self._validation_messages
    
    def get_errors(self) -> list[ValidationMessage]: 
        return filter_errors(self._validation_messages)
    
    def get_warnings(self) -> list[ValidationMessage]: 
        return filter_warnings(self._validation_messages)
    
    def is_valid(self) -> bool:
        return len(filter_errors(self.get_nested_validation_messages())) == 0

    # Function to extract warnings from a model and its nested models
    def get_nested_validation_messages(self, parent_name: str = "", visited: Set[int] = None) -> List[ValidationMessage]:
        """
        Recursively extract warnings from a Pydantic model and its nested fields.
        
        :param model: The Pydantic model instance to inspect.
        :param parent_name: The name of the parent model to track the path.
        :return: List of tuples containing (model name, warning message).
        """                   
        if visited is None:
            visited = set()

        model_id = id(self)
        if model_id in visited:
            return []
        visited.add(model_id)
        
        warnings_list = [warning for warning in self.get_validation_messages()]
        # warnings_list = [(parent_name or self.__class__.__name__, model_id,  warning) for warning in self.get_validation_messages()]


        for field_name, field in self.__fields__.items():
            full_path = f"{parent_name}.{field_name}" if parent_name else field_name
            value = getattr(self, field_name)

            if isinstance(value, BaseModelWithValidationMessages):
                warnings_list.extend(value.get_nested_validation_messages(full_path, visited))
            elif isinstance(value, list):
                for index, item in enumerate(value):
                    if isinstance(item, BaseModelWithValidationMessages):
                        list_path = f"{full_path}[{index}]"
                        warnings_list.extend(item.get_nested_validation_messages(list_path, visited))
        return warnings_list
    
    
    def get_nested_errors(self) -> list[ValidationMessage]: 
        return filter_errors(self.get_nested_validation_messages())
    
    def get_nested_warnings(self) -> list[ValidationMessage]: 
        return filter_warnings(self.get_nested_validation_messages())
    
    
    def str_for_validation_msg(self, validation_msg:ValidationMessage):
        if validation_msg.source_id == id(self):
            return validation_msg.source_id
            #return validation_msg.emphasize_in(self(str))
        else:
            return str(self)
        
    def str_highlighted(self):
        raise NotImplementedError("Subclasses must implement format_special()")
    
    
        
    def _emphasize_in(self, validation_msg, validation_node_str:str, fmt, color='black'):
        if validation_msg.highlight_sub_patterns:
            replacements = validation_msg.highlight_sub_patterns
        else:
            replacements = [validation_msg.highlight]
        # Sort patterns by length descending to avoid subpattern clobbering
        sorted_patterns = sorted(replacements, key=len, reverse=True)
        # Escape the patterns for regex safety
        escaped_patterns = [re.escape(p) for p in sorted_patterns]
        # Create one regex pattern with alternation (longest first)
        pattern = re.compile("|".join(escaped_patterns))
        
        out = pattern.sub(lambda m: fmt(m.group(0)), validation_node_str)
        return out
    
    
    def print_validation_messages(self, target='console'):
        msgs = self.get_nested_validation_messages()
        
        table = Table(title=f"Validation Results", show_header=False)

        col = lambda s:  table.add_column(s, vertical='top')
        col("-")
        
             
        if not msgs:
            table.add_row('All clear!', end_section=True)
            return

        for m in msgs:
            if m.level == ValidationMsgLevel.ERROR:
                color = 'red'
            else:
                color = 'yellow'
                
            match target:
                case 'markdown':
                    fmt = lambda s: f'🔸{s}🔸'
                case 'console':     
                    fmt = lambda s: f'[{color} bold]{s}[/{color} bold]'
                case 'html':
                    fmt = lambda s: f'<span class="val_{color}">{s}</span>'
                case 'html_styled':
                    fmt = lambda s: f'<b style="color:{color}>{s}</b>'
                
            serialized = str(self)
            emphazised_highlight = self._emphasize_in(m, serialized, fmt=fmt, color=color)
            
            txt =       f'[bold {color}]{m.level.name} [/bold {color}]'
            txt += '\n' + f'{m.problem_msg}'
            txt += '\n' + emphazised_highlight

            table.add_row( txt)
            table.add_section()

        print(table)
    
    # def print_validation_messages_(self, str_to_highlight_in=None, target='console'):
    #     if not str_to_highlight_in:
    #         str_to_highlight_in = str(self)
    #     msgs = self.get_nested_validation_messages()
    #     print('\n'.join(['\n',
    #                      '=======================================',
    #                      'Validation Results',
    #                      '---------------------------------------'
    #                     ]
    #                     )
    #     )
        
    #     if not msgs:
    #         print('All clear!')
    #         return

    #     for m in msgs:
    #         if m.level.casefold() == "error":
    #             color = 'red'
    #         else:
    #             color = 'yellow'
                
    #         text = Text.from_markup(f'\n [bold {color}]{m.level} [/bold {color}] in \t {m.source}' )
    #         print(text)
    #         match target:
    #             case 'markdown':
    #                 formatted_highlight = m.emphazised_highlight.replace('emph', f'🔸').replace('[/', '').replace('[', '').replace(']', '')
    #             case 'console':     
    #                 formatted_highlight = m.emphazised_highlight.replace('emph', f'bold {color}')
    #             case 'html':
    #                 formatted_highlight = m.emphazised_highlight.replace('emph', f'b').replace('[', '<').replace(']', '>')
    #         fmtd = str_to_highlight_in.replace(m.highlight, formatted_highlight)
    #         fmtd = Text.from_markup(fmtd)
    #         print(fmtd)
    #         print(Text.from_markup(f'{m.problem_msg}'))
        
    
    
def filter_errors(val_msg:list[ValidationMessage]) -> list[ValidationMessage]:
    return [ m for m in val_msg if m.level == ValidationMsgLevel.ERROR ]

def filter_warnings(val_msg:list[ValidationMessage]) -> list[ValidationMessage]:
    return [ m for m in val_msg if m.level != ValidationMsgLevel.ERROR  ]     




from pydantic import BaseModel

from labfreed.TREX.unece_units import unece_units           
                

class Unit(BaseModel):
    name: str
    symbol: str


class Quantity(BaseModel):
    value:int|float
    unit: Unit
      
    def __str__(self):
        unit_symbol = self.unit.symbol
        if unit_symbol == "dimensionless":
            unit_symbol = ""
        
        s = f"{str(self.value)} {unit_symbol}" 
        return s
    
    
def unece_unit_code_from_quantity(q:Quantity):
        by_name =   [ u['commonCode'] for u in unece_units() if u.get('name','') == q.unit.name] 
        by_symbol = [ u['commonCode'] for u in unece_units() if u.get('symbol','') == q.unit.symbol]
        code = list(set(by_name) | set(by_symbol))
        if len(code) != 1:
            raise ValueError(f'No UNECE unit code found for Quantity {str(q)}' ) 
        return code[0]
    
    
# class DataTable(list):
#     def __init__(self, headers:tuple[str, Any]):
#         for h in headers:
#             if len(h) != 2:
#                 raise ValueError(f'Headers must be tuples of length two. With a column name and type.')
#             if not isinstance(h[0], str):
#                 raise ValueError(f'Invalid type of header name {h[0]}. Must be str')
#             if not (h[1]):
#                 raise ValueError(f'Header type cannot be None')
#         self.headers = headers
#         super().__init__()
    
#     def append(self, row:list):
#         if len(row) != len(self.headers):
#             raise ValueError(f'Row has different length than headers')
#         super().append(row)

class DataTable(list):
    def __init__(self, col_names:list[str]=None):
        self.col_names = col_names
        self.row_template = None
        super().__init__()
    
    def append(self, row:list):
        if not self.row_template:
            self.row_template = row.copy()
        super().append(row)
        
    def extend(self, iterable):
        for item in iterable:
            self.append(item) 
        
        

         
if __name__ == "__main__":
    pass                
        
        

    


from labfreed.PAC_ID.data_model import PACID, IDSegment
from labfreed.PAC_ID.extensions import UnknownExtension
from labfreed.IO.parse_pac import PACID_With_Extensions


pac_id = PACID(issuer = 'mettorius.com',
            identifier = [  IDSegment(value='-DR'), 
                            IDSegment(value='999'),
                            IDSegment(value="-MD"),
                            IDSegment(key='240', value='1'),
                            IDSegment(key='21', value='2')
                        ]
          ) 

extensions= [
    UnknownExtension(name_= 'N', type_= 'T.T', data_= 'DATA'),
    UnknownExtension(name_= 'SUM', type_= 'TREX', data_= 'DATA'),
    UnknownExtension(name_= 'FOO', type_= 'T', data_= 'DATA'),
    UnknownExtension(name_= 'BAR', type_= 'T2', data_= 'DATA')
]

pac_with_ext = PACID_With_Extensions(pac_id=pac_id, extensions=extensions)


def test_url_serialization():
    url = pac_with_ext.serialize( use_short_notation_for_extensions=False, uppercase_only=True)
    assert url == 'HTTPS://PAC.METTORIUS.COM/-DR/999/-MD/240:1/21:2*N$T.T/DATA*SUM$TREX/DATA*FOO$T/DATA*BAR$T2/DATA'
    
def test_url_serialization_short_notation_for_N_and_SUM_extensions():
    url = pac_with_ext.serialize(use_short_notation_for_extensions=True, uppercase_only=True)
    assert url == 'HTTPS://PAC.METTORIUS.COM/-DR/999/-MD/240:1/21:2*DATA*DATA*FOO$T/DATA*BAR$T2/DATA'
    
def test_url_serialization_short_notation_for_N_extensions():
    ext = [extensions[i] for i in (0,2,3)]
    url = pac_with_ext.serialize( use_short_notation_for_extensions=True, uppercase_only=True)
    assert url == 'HTTPS://PAC.METTORIUS.COM/-DR/999/-MD/240:1/21:2*DATA*FOO$T/DATA*BAR$T2/DATA'
    
def test_url_serialization_short_notation_stop_implying_when_specific_extension():
    ext = [extensions[i] for i in (2,3)]
    url = pac_with_ext.serialize( use_short_notation_for_extensions=True, uppercase_only=True)
    assert url == 'HTTPS://PAC.METTORIUS.COM/-DR/999/-MD/240:1/21:2*FOO$T/DATA*BAR$T2/DATA'
    
    

    
    
    
    



from __future__ import annotations  # optional in 3.11, but recommended for consistency

from abc import ABC
from typing import Self
from pydantic import Field, computed_field, model_validator

from rich import print
from rich.text import Text
from rich.table import Table

from labfreed.validation import BaseModelWithValidationMessages, ValidationMsgLevel

from ..PAC_ID.data_model import PACID, IDSegment


    
class PAC_CAT(PACID):
    ''' 
    Extends a PAC-ID with interpretation of the identifier as categories
    '''
    categories:list[Category] = Field(default_factory=list)
    
    @property
    def identifier(self) -> list[IDSegment]:
        out = list()
        for category in self.categories:
            out.append(IDSegment(value=category.key))
            out.extend(category.segments)
        return out
    
    def get_category(self, key):
        tmp = [c for c in self.categories if c.key == key]
        if not tmp:
            return None
        return tmp[0]
    
    
    @classmethod
    def from_pac_id(cls, pac_id:PACID):
        d = pac_id.model_dump()
        issuer = d.get('issuer')
        
        category_segments = cls._split_into_categories(pac_id.identifier)
        categories = list()
        for c in category_segments:
            categories.append(cls.cat_from_cat_segments(c))
        
        return PAC_CAT(issuer=issuer, categories=categories, identifier=pac_id.identifier)
        
            
    @classmethod
    def cat_from_cat_segments(cls, segments:list[IDSegment]):
        segments = segments.copy()
        category_key = segments[0].value 
        segments.pop(0)
        
        mapping  = {
                '-MD': Material_Device,
                '-MS': Material_Substance,
                '-MC': Material_Consumable,
                '-MM': Material_Misc,
                '-DM': Data_Method,
                '-DR': Data_Result,
                '-DC': Data_Calibration,
                '-DP': Data_Progress,
                '-DS': Data_Static 
        }
        cat = mapping.get(category_key) or Category

        # implicit segment keys
        model_dict = {v.alias: None for k, v in cat.model_fields.items() if v.alias and k not in ['key','additional_segments']}
        for k, seg in zip(model_dict.keys(), segments.copy()):
            if seg.key:
                break
            model_dict[k] = seg.value
            segments.pop(0)
            
        # try to fill model keys if not already set
        for s in segments:
            if s.key in model_dict and not model_dict.get(s.key):
                model_dict[s.key] = s.value
                segments.remove(s)
            
        model_dict['additional_segments'] = segments
        model_dict['key'] = category_key
        return cat(**model_dict)
    
    @staticmethod
    def _split_into_categories(segments:list[IDSegment]):
        categories = list()
        c_segments = list()
        categories.append(c_segments)
        for s in segments:
            # new category starts with "-"
            if s.value[0] == '-':
                cat_key = s.value
                c = [s]
                categories.append(c)
            else:
                c.append(s)
                
        # first cat can be empty > remove
        categories = [c for c in categories if len(c) > 0]
                            
        return categories
    
            
    # @computed_field
    # @property
    # def categories(self) -> list[Category]:
    #     categories = list()
    #     c = Category(segments=[])
    #     categories.append(c)
    #     for s in self.identifier:
    #         s:IDSegment = s
    #         # new category starts with "-"
    #         if s.value[0] == '-':
    #             cat_key = s.value
    #             c = Category(key=cat_key, segments=[])
    #             categories.append(c)
    #         else:
    #             c.segments.append(s)
            
    #     # the first category might have no segments. remove categories without segments
    #     if not categories[0].segments:
    #         categories = categories[1:]
                
    #     return categories
    
    @model_validator(mode='after')
    def check_keys_are_unique_in_each_category(self) -> Self:
        for c in self.categories:
            keys = [s.key for s in c.segments if s.key]
            duplicate_keys = [k for k in set(keys) if keys.count(k) > 1]
            if duplicate_keys:
                for k in duplicate_keys:
                    self.add_validation_message(
                        source=f"identifier {k}",
                        level = ValidationMsgLevel.ERROR,
                        msg=f"Duplicate key {k} in category {c.key}",
                        highlight_pattern = k
                    )
            return self
        
    def print_categories(self):
        
        table = Table(title=f'Categories in {str(self)}', show_header=False)
        table.add_column('0')
        table.add_column('1')
        s = ''
        for i, c in enumerate(self.categories):
            if i == 0:
                title = Text('Main Category', style='bold')
            else:
                title = Text('Category', style='bold')
            
            table.add_row(title)
            
            for field_name, field_info in c.model_fields.items():
                if not getattr(c, field_name):
                    continue
                table.add_row(f"{field_name} ({field_info.alias or ''})",
                              f" {getattr(c, field_name)}"
                              )      
            table.add_section()        
        print(table)
        
 

        
        
    # @classmethod
    # def from_categories(cls, issuer:str, categories:list[Category]):
    #     return PAC_CAT(issuer=issuer, identifier=cls.identifier_from_categories(categories))
        
       
    # @classmethod 
    # def identifier_from_categories(cls, categories:list[Category]) :
    #     segments = list()
    #     for c in categories:
    #         if c.key:
    #             segments.append(IDSegment(value=c.key))
    #         segments.extend(c.segments)
    #     return segments
    
    

                

class Category(BaseModelWithValidationMessages):
    model_config = {
        "populate_by_name": True
    }
    key:str
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)   
    
    @computed_field
    @property
    def segments(self, use_short_notation=False) -> list[IDSegment]:
        segments = []
        can_omit_keys = use_short_notation # keeps track of whether keys can still be omitted. That is the case when the segment recommendation is followed
        for field_name, field_info in self.model_fields.items():
            if field_name in ['key', 'additional_segments']:
                continue
            if value := getattr(self, field_name):
                if can_omit_keys:
                    key = None
                else:
                    key = field_info.alias
                segments.append(IDSegment(key= key, value= value)  )
            else:
                can_omit_keys = False
        if self.additional_segments:     
            segments.extend(self.additional_segments)
            
        return segments

    
    @model_validator(mode='after')
    def warn_unusual_category_key(self):
        ''' this base class is instantiated only if the key is not a known category key'''
        if type(self) is Category:
            self.add_validation_message(
                        source=f"Category {self.key}",
                        level = ValidationMsgLevel.RECOMMENDATION,
                        msg=f'Category key {self.key} is not a well known key. It is recommended to use well known keys only',
                        highlight_pattern = f"{self.key}"
            )
        return self
    
    
    def __str__(self):
        s = '\n'.join( [f'{field_name} \t ({field_info.alias or ''}): \t {getattr(self, field_name)}' for  field_name, field_info in self.model_fields.items() if getattr(self, field_name)]) 
        return s
 

    
    
    # def to_identifier_category(self, use_short_notation=False):
    #     '''Creates a Category with the correct segments. 
    #        Segments are in order of the Pydantic model fields. 
    #        Segment keys are omitted as long as the recommendation is followed.
    #        Additional segments are added at the end'''
    #     segments = []
    #     can_omit_keys = use_short_notation # keeps track of whether keys can still be omitted. That is the case when the segment recommendation is followed
    #     for field_name, field_info in self.model_fields.items():
    #         if field_name in ['category_key', 'additional_segments']:
    #             continue
    #         if value := getattr(self, field_name):
    #             if can_omit_keys:
    #                 key = None
    #             else:
    #                 key = field_info.alias
    #             segments.append(IDSegment(key= key, value= value)  )
    #         else:
    #             can_omit_keys = False
    #     if self.additional_segments:     
    #         segments.extend(self.additional_segments)
    #     return Category(key=self.key,
    #                     segments=segments)
        
        
        
# def _apply_category_defaults(self, segments_in: list[IDSegment]):
        
#         category_conventions = MappingProxyType(
#                             {
#                                 '-MD': ['240', '21'],
#                                 '-MS': ['240', '10', '20', '21', '250'],
#                                 '-MC': ['240', '10', '20', '21', '250'],
#                                 '-MM': ['240', '10', '20', '21', '250']
#                             }
#                         )
        
#         segments = segments_in.copy()
#         default_keys = None
#         for s in segments:
#             if not s.key and default_keys:
#                 s.key = default_keys.pop(0)
#             else:
#                 default_keys = None
                
#             # category starts: start with new defaults. 
#             if s.value in category_conventions.keys():
#                 default_keys = category_conventions.get(s.value).copy() #copy, so the entries can be popped when used
#         return segments
                
            


class Material_Device(Category):
    key: str =         Field(default='-MD', frozen=True)
    model_number: str|None =         Field(              alias='240')
    serial_number: str|None =        Field(              alias='21')
    
    @model_validator(mode='after')
    def validate_mandatory_fields(self):
        if not self.model_number:
            self.add_validation_message(
                    source=f"Category {self.key}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f'Category key {self.key} is missing mandatory field Model Number',
                    highlight_pattern = f"{self.key}"
            )
        if not self.serial_number:
            self.add_validation_message(
                    source=f"Category {self.key}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f'Category key {self.key} is missing mandatory field Serial Number',
                    highlight_pattern = f"{self.key}"
            )
    
class Material_Substance(Category):
    key: str =         Field(default='-MS', frozen=True)
    product_number:str|None =    Field(              alias='240')
    batch_number:str|None =     Field(default=None, alias='10')
    container_size:str|None =   Field(default=None, alias='20')
    container_number:str|None = Field(default=None, alias='21')
    aliquot:str|None =          Field(default=None, alias='250')
    
    @model_validator(mode='after')
    def validate_mandatory_fields(self):
        if not self.product_number:
            self.add_validation_message(
                    source=f"Category {self.key}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f'Category key {self.key} is missing mandatory field Product Number',
                    highlight_pattern = f"{self.key}"
            )
    
class Material_Consumable(Category):
    key: str = Field(default='-MC', frozen=True)
    product_number:str|None =   Field(              alias='240')
    batch_number:str|None =     Field(default=None, alias='10')
    packing_size:str|None =     Field(default=None, alias='20')
    serial_number:str|None =    Field(default=None, alias='21')
    aliquot:str|None =          Field(default=None, alias='250')
    
    @model_validator(mode='after')
    def validate_mandatory_fields(self):
        if not self.product_number:
            self.add_validation_message(
                    source=f"Category {self.key}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f"Category key {self.key} is missing mandatory field 'Product Number'",
                    highlight_pattern = f"{self.key}"
            )
    
class Material_Misc(Material_Consumable):
    key: str = Field(default='-MM', frozen=True)
    



class Data_Abstract(Category, ABC):
    key: str
    id:str|None =                    Field(              alias='21')
    
    @model_validator(mode='after')
    def validate_mandatory_fields(self):
        if not self.id:
            self.add_validation_message(
                    source=f"Category {self.key}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f"Category key {self.key} is missing mandatory field 'ID'",
                    highlight_pattern = f"{self.key}"
            )

class Data_Result(Data_Abstract):
    key: str = Field(default='-DR', frozen=True)
    
class Data_Method(Data_Abstract):
    key: str = Field(default='-DM', frozen=True)
    
class Data_Calibration(Data_Abstract):
    key: str = Field(default='-DC', frozen=True)
    
class Data_Progress(Data_Abstract):
    key: str = Field(default='-DP', frozen=True)
    
class Data_Static(Data_Abstract):
    key: str = Field(default='-DS', frozen=True)
    
    
    
    
# mapping  = {
#         '-MD': Material_Device,
#         '-MS': Material_Substance,
#         '-MC': Material_Consumable,
#         '-MM': Material_Misc,
#         '-DM': Data_Method,
#         '-DR': Data_Result,
#         '-DC': Data_Calibration,
#         '-DP': Data_Progress,
#         '-DS': Data_Static 
#     }
    
# def CAT_from_category(category:Category) -> Category|None:
#     raise NotImplementedError()

# def CAT_from_category_key(category_key) -> Category|None:
#     return mapping.get(category_key)
    

# if __name__ == "__main__":
#     pass